/*
 * Copyright (c) 2018-2019, FusionAuth, All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific
 * language governing permissions and limitations under the License.
 */

'use strict';

var FusionAuthClient = function(apiKey, host) {
  this.apiKey = apiKey;
  this.host = host;
  this.tenantId = null;
};

FusionAuthClient.constructor = FusionAuthClient;
//noinspection JSUnusedGlobalSymbols
FusionAuthClient.prototype = {

  /**
   * Sets the tenantId on the client.
   *
   * @param tenantId
   */
  setTenantId: function(tenantId) {
    this.tenantId = tenantId;
    return this;
  },

  /**
   * Takes an action on a user. The user being actioned is called the "actionee" and the user taking the action is called the
   * "actioner". Both user ids are required in the request object.
   *
   * @param {ActionRequest} request The action request that includes all of the information about the action being taken including
   *    the id of the action, any options and the duration (if applicable).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  actionUser: function(request, callBack) {
      return this._start()
          .uri('/api/user/action')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Activates the FusionAuth Reactor using a license id and optionally a license text (for air-gapped deployments)
   *
   * @param {ReactorRequest} request An optional request that contains the license text to activate Reactor (useful for air-gap deployments of FusionAuth).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  activateReactor: function(request, callBack) {
      return this._start()
          .uri('/api/reactor')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Adds a user to an existing family. The family id must be specified.
   *
   * @param {UUIDString} familyId The id of the family.
   * @param {FamilyRequest} request The request object that contains all of the information used to determine which user to add to the family.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  addUserToFamily: function(familyId, request, callBack) {
      return this._start()
          .uri('/api/user/family')
          .urlSegment(familyId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Cancels the user action.
   *
   * @param {UUIDString} actionId The action id of the action to cancel.
   * @param {ActionRequest} request The action request that contains the information about the cancellation.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  cancelAction: function(actionId, request, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlSegment(actionId)
          .setJSONBody(request)
          .delete()
          .go(callBack);
  },

  /**
   * Changes a user's password using the change password Id. This usually occurs after an email has been sent to the user
   * and they clicked on a link to reset their password.
   *
   * @param {string} changePasswordId The change password Id used to find the user. This value is generated by FusionAuth once the change password workflow has been initiated.
   * @param {ChangePasswordRequest} request The change password request that contains all of the information used to change the password.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  changePassword: function(changePasswordId, request, callBack) {
      return this._startAnonymous()
          .uri('/api/user/change-password')
          .urlSegment(changePasswordId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Changes a user's password using their identity (login id and password). Using a loginId instead of the changePasswordId
   * bypasses the email verification and allows a password to be changed directly without first calling the #forgotPassword
   * method.
   *
   * @param {ChangePasswordRequest} request The change password request that contains all of the information used to change the password.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  changePasswordByIdentity: function(request, callBack) {
      return this._start()
          .uri('/api/user/change-password')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Adds a comment to the user's account.
   *
   * @param {UserCommentRequest} request The request object that contains all of the information used to create the user comment.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  commentOnUser: function(request, callBack) {
      return this._start()
          .uri('/api/user/comment')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an API key. You can optionally specify a unique Id for the key, if not provided one will be generated.
   * an API key can only be created with equal or lesser authority. An API key cannot create another API key unless it is granted 
   * to that API key.
   * 
   * If an API key is locked to a tenant, it can only create API Keys for that same tenant.
   *
   * @param {?UUIDString} keyId (Optional) The unique Id of the API key. If not provided a secure random Id will be generated.
   * @param {APIKeyRequest} request The request object that contains all of the information needed to create the APIKey.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createAPIKey: function(keyId, request, callBack) {
      return this._start()
          .uri('/api/api-key')
          .urlSegment(keyId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an application. You can optionally specify an Id for the application, if not provided one will be generated.
   *
   * @param {?UUIDString} applicationId (Optional) The Id to use for the application. If not provided a secure random UUID will be generated.
   * @param {ApplicationRequest} request The request object that contains all of the information used to create the application.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createApplication: function(applicationId, request, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a new role for an application. You must specify the id of the application you are creating the role for.
   * You can optionally specify an Id for the role inside the ApplicationRole object itself, if not provided one will be generated.
   *
   * @param {UUIDString} applicationId The Id of the application to create the role on.
   * @param {?UUIDString} roleId (Optional) The Id of the role. If not provided a secure random UUID will be generated.
   * @param {ApplicationRequest} request The request object that contains all of the information used to create the application role.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createApplicationRole: function(applicationId, roleId, request, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment("role")
          .urlSegment(roleId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an audit log with the message and user name (usually an email). Audit logs should be written anytime you
   * make changes to the FusionAuth database. When using the FusionAuth App web interface, any changes are automatically
   * written to the audit log. However, if you are accessing the API, you must write the audit logs yourself.
   *
   * @param {AuditLogRequest} request The request object that contains all of the information used to create the audit log entry.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createAuditLog: function(request, callBack) {
      return this._start()
          .uri('/api/system/audit-log')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a connector.  You can optionally specify an Id for the connector, if not provided one will be generated.
   *
   * @param {?UUIDString} connectorId (Optional) The Id for the connector. If not provided a secure random UUID will be generated.
   * @param {ConnectorRequest} request The request object that contains all of the information used to create the connector.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createConnector: function(connectorId, request, callBack) {
      return this._start()
          .uri('/api/connector')
          .urlSegment(connectorId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a user consent type. You can optionally specify an Id for the consent type, if not provided one will be generated.
   *
   * @param {?UUIDString} consentId (Optional) The Id for the consent. If not provided a secure random UUID will be generated.
   * @param {ConsentRequest} request The request object that contains all of the information used to create the consent.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createConsent: function(consentId, request, callBack) {
      return this._start()
          .uri('/api/consent')
          .urlSegment(consentId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an email template. You can optionally specify an Id for the template, if not provided one will be generated.
   *
   * @param {?UUIDString} emailTemplateId (Optional) The Id for the template. If not provided a secure random UUID will be generated.
   * @param {EmailTemplateRequest} request The request object that contains all of the information used to create the email template.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createEmailTemplate: function(emailTemplateId, request, callBack) {
      return this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an Entity. You can optionally specify an Id for the Entity. If not provided one will be generated.
   *
   * @param {?UUIDString} entityId (Optional) The Id for the Entity. If not provided a secure random UUID will be generated.
   * @param {EntityRequest} request The request object that contains all of the information used to create the Entity.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createEntity: function(entityId, request, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a Entity Type. You can optionally specify an Id for the Entity Type, if not provided one will be generated.
   *
   * @param {?UUIDString} entityTypeId (Optional) The Id for the Entity Type. If not provided a secure random UUID will be generated.
   * @param {EntityTypeRequest} request The request object that contains all of the information used to create the Entity Type.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createEntityType: function(entityTypeId, request, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a new permission for an entity type. You must specify the id of the entity type you are creating the permission for.
   * You can optionally specify an Id for the permission inside the EntityTypePermission object itself, if not provided one will be generated.
   *
   * @param {UUIDString} entityTypeId The Id of the entity type to create the permission on.
   * @param {?UUIDString} permissionId (Optional) The Id of the permission. If not provided a secure random UUID will be generated.
   * @param {EntityTypeRequest} request The request object that contains all of the information used to create the permission.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createEntityTypePermission: function(entityTypeId, permissionId, request, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .urlSegment("permission")
          .urlSegment(permissionId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a family with the user id in the request as the owner and sole member of the family. You can optionally specify an id for the
   * family, if not provided one will be generated.
   *
   * @param {?UUIDString} familyId (Optional) The id for the family. If not provided a secure random UUID will be generated.
   * @param {FamilyRequest} request The request object that contains all of the information used to create the family.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createFamily: function(familyId, request, callBack) {
      return this._start()
          .uri('/api/user/family')
          .urlSegment(familyId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a form.  You can optionally specify an Id for the form, if not provided one will be generated.
   *
   * @param {?UUIDString} formId (Optional) The Id for the form. If not provided a secure random UUID will be generated.
   * @param {FormRequest} request The request object that contains all of the information used to create the form.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createForm: function(formId, request, callBack) {
      return this._start()
          .uri('/api/form')
          .urlSegment(formId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a form field.  You can optionally specify an Id for the form, if not provided one will be generated.
   *
   * @param {?UUIDString} fieldId (Optional) The Id for the form field. If not provided a secure random UUID will be generated.
   * @param {FormFieldRequest} request The request object that contains all of the information used to create the form field.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createFormField: function(fieldId, request, callBack) {
      return this._start()
          .uri('/api/form/field')
          .urlSegment(fieldId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a group. You can optionally specify an Id for the group, if not provided one will be generated.
   *
   * @param {?UUIDString} groupId (Optional) The Id for the group. If not provided a secure random UUID will be generated.
   * @param {GroupRequest} request The request object that contains all of the information used to create the group.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createGroup: function(groupId, request, callBack) {
      return this._start()
          .uri('/api/group')
          .urlSegment(groupId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a member in a group.
   *
   * @param {MemberRequest} request The request object that contains all of the information used to create the group member(s).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createGroupMembers: function(request, callBack) {
      return this._start()
          .uri('/api/group/member')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an identity provider. You can optionally specify an Id for the identity provider, if not provided one will be generated.
   *
   * @param {?UUIDString} identityProviderId (Optional) The Id of the identity provider. If not provided a secure random UUID will be generated.
   * @param {IdentityProviderRequest} request The request object that contains all of the information used to create the identity provider.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createIdentityProvider: function(identityProviderId, request, callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .urlSegment(identityProviderId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a Lambda. You can optionally specify an Id for the lambda, if not provided one will be generated.
   *
   * @param {?UUIDString} lambdaId (Optional) The Id for the lambda. If not provided a secure random UUID will be generated.
   * @param {LambdaRequest} request The request object that contains all of the information used to create the lambda.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createLambda: function(lambdaId, request, callBack) {
      return this._start()
          .uri('/api/lambda')
          .urlSegment(lambdaId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates an message template. You can optionally specify an Id for the template, if not provided one will be generated.
   *
   * @param {?UUIDString} messageTemplateId (Optional) The Id for the template. If not provided a secure random UUID will be generated.
   * @param {MessageTemplateRequest} request The request object that contains all of the information used to create the message template.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createMessageTemplate: function(messageTemplateId, request, callBack) {
      return this._start()
          .uri('/api/message/template')
          .urlSegment(messageTemplateId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a messenger.  You can optionally specify an Id for the messenger, if not provided one will be generated.
   *
   * @param {?UUIDString} messengerId (Optional) The Id for the messenger. If not provided a secure random UUID will be generated.
   * @param {MessengerRequest} request The request object that contains all of the information used to create the messenger.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createMessenger: function(messengerId, request, callBack) {
      return this._start()
          .uri('/api/messenger')
          .urlSegment(messengerId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a tenant. You can optionally specify an Id for the tenant, if not provided one will be generated.
   *
   * @param {?UUIDString} tenantId (Optional) The Id for the tenant. If not provided a secure random UUID will be generated.
   * @param {TenantRequest} request The request object that contains all of the information used to create the tenant.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createTenant: function(tenantId, request, callBack) {
      return this._start()
          .uri('/api/tenant')
          .urlSegment(tenantId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a Theme. You can optionally specify an Id for the theme, if not provided one will be generated.
   *
   * @param {?UUIDString} themeId (Optional) The Id for the theme. If not provided a secure random UUID will be generated.
   * @param {ThemeRequest} request The request object that contains all of the information used to create the theme.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createTheme: function(themeId, request, callBack) {
      return this._start()
          .uri('/api/theme')
          .urlSegment(themeId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a user. You can optionally specify an Id for the user, if not provided one will be generated.
   *
   * @param {?UUIDString} userId (Optional) The Id for the user. If not provided a secure random UUID will be generated.
   * @param {UserRequest} request The request object that contains all of the information used to create the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createUser: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a user action. This action cannot be taken on a user until this call successfully returns. Anytime after
   * that the user action can be applied to any user.
   *
   * @param {?UUIDString} userActionId (Optional) The Id for the user action. If not provided a secure random UUID will be generated.
   * @param {UserActionRequest} request The request object that contains all of the information used to create the user action.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createUserAction: function(userActionId, request, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a user reason. This user action reason cannot be used when actioning a user until this call completes
   * successfully. Anytime after that the user action reason can be used.
   *
   * @param {?UUIDString} userActionReasonId (Optional) The Id for the user action reason. If not provided a secure random UUID will be generated.
   * @param {UserActionReasonRequest} request The request object that contains all of the information used to create the user action reason.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createUserActionReason: function(userActionReasonId, request, callBack) {
      return this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a single User consent.
   *
   * @param {?UUIDString} userConsentId (Optional) The Id for the User consent. If not provided a secure random UUID will be generated.
   * @param {UserConsentRequest} request The request that contains the user consent information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createUserConsent: function(userConsentId, request, callBack) {
      return this._start()
          .uri('/api/user/consent')
          .urlSegment(userConsentId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Link an external user from a 3rd party identity provider to a FusionAuth user.
   *
   * @param {IdentityProviderLinkRequest} request The request object that contains all of the information used to link the FusionAuth user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createUserLink: function(request, callBack) {
      return this._start()
          .uri('/api/identity-provider/link')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Creates a webhook. You can optionally specify an Id for the webhook, if not provided one will be generated.
   *
   * @param {?UUIDString} webhookId (Optional) The Id for the webhook. If not provided a secure random UUID will be generated.
   * @param {WebhookRequest} request The request object that contains all of the information used to create the webhook.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  createWebhook: function(webhookId, request, callBack) {
      return this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Deactivates the application with the given Id.
   *
   * @param {UUIDString} applicationId The Id of the application to deactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deactivateApplication: function(applicationId, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .delete()
          .go(callBack);
  },

  /**
   * Deactivates the FusionAuth Reactor.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deactivateReactor: function(callBack) {
      return this._start()
          .uri('/api/reactor')
          .delete()
          .go(callBack);
  },

  /**
   * Deactivates the user with the given Id.
   *
   * @param {UUIDString} userId The Id of the user to deactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deactivateUser: function(userId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .delete()
          .go(callBack);
  },

  /**
   * Deactivates the user action with the given Id.
   *
   * @param {UUIDString} userActionId The Id of the user action to deactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deactivateUserAction: function(userActionId, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .delete()
          .go(callBack);
  },

  /**
   * Deactivates the users with the given ids.
   *
   * @param {Array<string>} userIds The ids of the users to deactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to deactivateUsersByIds, use that method instead.
   */
  deactivateUsers: function(userIds, callBack) {
      return this._start()
          .uri('/api/user/bulk')
          .urlParameter('userId', userIds)
          .urlParameter('dryRun', false)
          .urlParameter('hardDelete', false)
          .delete()
          .go(callBack);
  },

  /**
   * Deactivates the users with the given ids.
   *
   * @param {Array<string>} userIds The ids of the users to deactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deactivateUsersByIds: function(userIds, callBack) {
      return this._start()
          .uri('/api/user/bulk')
          .urlParameter('userId', userIds)
          .urlParameter('dryRun', false)
          .urlParameter('hardDelete', false)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the API key for the given Id.
   *
   * @param {UUIDString} keyId The Id of the authentication API key to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteAPIKey: function(keyId, callBack) {
      return this._start()
          .uri('/api/api-key')
          .urlSegment(keyId)
          .delete()
          .go(callBack);
  },

  /**
   * Hard deletes an application. This is a dangerous operation and should not be used in most circumstances. This will
   * delete the application, any registrations for that application, metrics and reports for the application, all the
   * roles for the application, and any other data associated with the application. This operation could take a very
   * long time, depending on the amount of data in your database.
   *
   * @param {UUIDString} applicationId The Id of the application to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteApplication: function(applicationId, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlParameter('hardDelete', true)
          .delete()
          .go(callBack);
  },

  /**
   * Hard deletes an application role. This is a dangerous operation and should not be used in most circumstances. This
   * permanently removes the given role from all users that had it.
   *
   * @param {UUIDString} applicationId The Id of the application to deactivate.
   * @param {UUIDString} roleId The Id of the role to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteApplicationRole: function(applicationId, roleId, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment("role")
          .urlSegment(roleId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the connector for the given Id.
   *
   * @param {UUIDString} connectorId The Id of the connector to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteConnector: function(connectorId, callBack) {
      return this._start()
          .uri('/api/connector')
          .urlSegment(connectorId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the consent for the given Id.
   *
   * @param {UUIDString} consentId The Id of the consent to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteConsent: function(consentId, callBack) {
      return this._start()
          .uri('/api/consent')
          .urlSegment(consentId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the email template for the given Id.
   *
   * @param {UUIDString} emailTemplateId The Id of the email template to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteEmailTemplate: function(emailTemplateId, callBack) {
      return this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the Entity for the given Id.
   *
   * @param {UUIDString} entityId The Id of the Entity to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteEntity: function(entityId, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes an Entity Grant for the given User or Entity.
   *
   * @param {UUIDString} entityId The Id of the Entity that the Entity Grant is being deleted for.
   * @param {?UUIDString} recipientEntityId (Optional) The Id of the Entity that the Entity Grant is for.
   * @param {?UUIDString} userId (Optional) The Id of the User that the Entity Grant is for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteEntityGrant: function(entityId, recipientEntityId, userId, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .urlSegment("grant")
          .urlParameter('recipientEntityId', recipientEntityId)
          .urlParameter('userId', userId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the Entity Type for the given Id.
   *
   * @param {UUIDString} entityTypeId The Id of the Entity Type to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteEntityType: function(entityTypeId, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .delete()
          .go(callBack);
  },

  /**
   * Hard deletes a permission. This is a dangerous operation and should not be used in most circumstances. This
   * permanently removes the given permission from all grants that had it.
   *
   * @param {UUIDString} entityTypeId The Id of the entityType the the permission belongs to.
   * @param {UUIDString} permissionId The Id of the permission to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteEntityTypePermission: function(entityTypeId, permissionId, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .urlSegment("permission")
          .urlSegment(permissionId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the form for the given Id.
   *
   * @param {UUIDString} formId The Id of the form to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteForm: function(formId, callBack) {
      return this._start()
          .uri('/api/form')
          .urlSegment(formId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the form field for the given Id.
   *
   * @param {UUIDString} fieldId The Id of the form field to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteFormField: function(fieldId, callBack) {
      return this._start()
          .uri('/api/form/field')
          .urlSegment(fieldId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the group for the given Id.
   *
   * @param {UUIDString} groupId The Id of the group to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteGroup: function(groupId, callBack) {
      return this._start()
          .uri('/api/group')
          .urlSegment(groupId)
          .delete()
          .go(callBack);
  },

  /**
   * Removes users as members of a group.
   *
   * @param {MemberDeleteRequest} request The member request that contains all of the information used to remove members to the group.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteGroupMembers: function(request, callBack) {
      return this._start()
          .uri('/api/group/member')
          .setJSONBody(request)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the identity provider for the given Id.
   *
   * @param {UUIDString} identityProviderId The Id of the identity provider to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteIdentityProvider: function(identityProviderId, callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .urlSegment(identityProviderId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the key for the given Id.
   *
   * @param {UUIDString} keyId The Id of the key to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteKey: function(keyId, callBack) {
      return this._start()
          .uri('/api/key')
          .urlSegment(keyId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the lambda for the given Id.
   *
   * @param {UUIDString} lambdaId The Id of the lambda to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteLambda: function(lambdaId, callBack) {
      return this._start()
          .uri('/api/lambda')
          .urlSegment(lambdaId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the message template for the given Id.
   *
   * @param {UUIDString} messageTemplateId The Id of the message template to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteMessageTemplate: function(messageTemplateId, callBack) {
      return this._start()
          .uri('/api/message/template')
          .urlSegment(messageTemplateId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the messenger for the given Id.
   *
   * @param {UUIDString} messengerId The Id of the messenger to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteMessenger: function(messengerId, callBack) {
      return this._start()
          .uri('/api/messenger')
          .urlSegment(messengerId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the user registration for the given user and application.
   *
   * @param {UUIDString} userId The Id of the user whose registration is being deleted.
   * @param {UUIDString} applicationId The Id of the application to remove the registration for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteRegistration: function(userId, applicationId, callBack) {
      return this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .urlSegment(applicationId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the tenant for the given Id.
   *
   * @param {UUIDString} tenantId The Id of the tenant to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteTenant: function(tenantId, callBack) {
      return this._start()
          .uri('/api/tenant')
          .urlSegment(tenantId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the tenant for the given Id asynchronously.
   * This method is helpful if you do not want to wait for the delete operation to complete.
   *
   * @param {UUIDString} tenantId The Id of the tenant to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteTenantAsync: function(tenantId, callBack) {
      return this._start()
          .uri('/api/tenant')
          .urlSegment(tenantId)
          .urlParameter('async', true)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the theme for the given Id.
   *
   * @param {UUIDString} themeId The Id of the theme to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteTheme: function(themeId, callBack) {
      return this._start()
          .uri('/api/theme')
          .urlSegment(themeId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the user for the given Id. This permanently deletes all information, metrics, reports and data associated
   * with the user.
   *
   * @param {UUIDString} userId The Id of the user to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteUser: function(userId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .urlParameter('hardDelete', true)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the user action for the given Id. This permanently deletes the user action and also any history and logs of
   * the action being applied to any users.
   *
   * @param {UUIDString} userActionId The Id of the user action to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteUserAction: function(userActionId, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .urlParameter('hardDelete', true)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the user action reason for the given Id.
   *
   * @param {UUIDString} userActionReasonId The Id of the user action reason to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteUserActionReason: function(userActionReasonId, callBack) {
      return this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .delete()
          .go(callBack);
  },

  /**
   * Remove an existing link that has been made from a 3rd party identity provider to a FusionAuth user.
   *
   * @param {UUIDString} identityProviderId The unique Id of the identity provider.
   * @param {string} identityProviderUserId The unique Id of the user in the 3rd party identity provider to unlink.
   * @param {UUIDString} userId The unique Id of the FusionAuth user to unlink.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteUserLink: function(identityProviderId, identityProviderUserId, userId, callBack) {
      return this._start()
          .uri('/api/identity-provider/link')
          .urlParameter('identityProviderId', identityProviderId)
          .urlParameter('identityProviderUserId', identityProviderUserId)
          .urlParameter('userId', userId)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the users with the given ids, or users matching the provided JSON query or queryString.
   * The order of preference is ids, query and then queryString, it is recommended to only provide one of the three for the request.
   * 
   * This method can be used to deactivate or permanently delete (hard-delete) users based upon the hardDelete boolean in the request body.
   * Using the dryRun parameter you may also request the result of the action without actually deleting or deactivating any users.
   *
   * @param {UserDeleteRequest} request The UserDeleteRequest.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to deleteUsersByQuery, use that method instead.
   */
  deleteUsers: function(request, callBack) {
      return this._start()
          .uri('/api/user/bulk')
          .setJSONBody(request)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the users with the given ids, or users matching the provided JSON query or queryString.
   * The order of preference is ids, query and then queryString, it is recommended to only provide one of the three for the request.
   * 
   * This method can be used to deactivate or permanently delete (hard-delete) users based upon the hardDelete boolean in the request body.
   * Using the dryRun parameter you may also request the result of the action without actually deleting or deactivating any users.
   *
   * @param {UserDeleteRequest} request The UserDeleteRequest.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteUsersByQuery: function(request, callBack) {
      return this._start()
          .uri('/api/user/bulk')
          .setJSONBody(request)
          .delete()
          .go(callBack);
  },

  /**
   * Deletes the webhook for the given Id.
   *
   * @param {UUIDString} webhookId The Id of the webhook to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  deleteWebhook: function(webhookId, callBack) {
      return this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .delete()
          .go(callBack);
  },

  /**
   * Disable Two Factor authentication for a user.
   *
   * @param {UUIDString} userId The Id of the User for which you're disabling Two Factor authentication.
   * @param {string} methodId The two-factor method identifier you wish to disable
   * @param {string} code The Two Factor code used verify the the caller knows the Two Factor secret.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  disableTwoFactor: function(userId, methodId, code, callBack) {
      return this._start()
          .uri('/api/user/two-factor')
          .urlParameter('userId', userId)
          .urlParameter('methodId', methodId)
          .urlParameter('code', code)
          .delete()
          .go(callBack);
  },

  /**
   * Enable Two Factor authentication for a user.
   *
   * @param {UUIDString} userId The Id of the user to enable Two Factor authentication.
   * @param {TwoFactorRequest} request The two factor enable request information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  enableTwoFactor: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user/two-factor')
          .urlSegment(userId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Exchanges an OAuth authorization code for an access token.
   * Makes a request to the Token endpoint to exchange the authorization code returned from the Authorize endpoint for an access token.
   *
   * @param {string} code The authorization code returned on the /oauth2/authorize response.
   * @param {string} client_id The unique client identifier. The client Id is the Id of the FusionAuth Application in which you you are attempting to authenticate.
   * @param {?string} client_secret (Optional) The client secret. This value will be required if client authentication is enabled.
   * @param {string} redirect_uri The URI to redirect to upon a successful request.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  exchangeOAuthCodeForAccessToken: function(code, client_id, client_secret, redirect_uri, callBack) {
      var body = new FormData();
      body.append('code', code);
      body.append('client_id', client_id);
      body.append('client_secret', client_secret);
      body.append('grant_type', 'authorization_code');
      body.append('redirect_uri', redirect_uri);
      return this._startAnonymous()
          .uri('/oauth2/token')
          .setFormData(body)
          .post()
          .go(callBack);
  },

  /**
   * Exchanges an OAuth authorization code and code_verifier for an access token.
   * Makes a request to the Token endpoint to exchange the authorization code returned from the Authorize endpoint and a code_verifier for an access token.
   *
   * @param {string} code The authorization code returned on the /oauth2/authorize response.
   * @param {?string} client_id (Optional) The unique client identifier. The client Id is the Id of the FusionAuth Application in which you you are attempting to authenticate. This parameter is optional when the Authorization header is provided.
   * @param {?string} client_secret (Optional) The client secret. This value may optionally be provided in the request body instead of the Authorization header.
   * @param {string} redirect_uri The URI to redirect to upon a successful request.
   * @param {string} code_verifier The random string generated previously. Will be compared with the code_challenge sent previously, which allows the OAuth provider to authenticate your app.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  exchangeOAuthCodeForAccessTokenUsingPKCE: function(code, client_id, client_secret, redirect_uri, code_verifier, callBack) {
      var body = new FormData();
      body.append('code', code);
      body.append('client_id', client_id);
      body.append('client_secret', client_secret);
      body.append('grant_type', 'authorization_code');
      body.append('redirect_uri', redirect_uri);
      body.append('code_verifier', code_verifier);
      return this._startAnonymous()
          .uri('/oauth2/token')
          .setFormData(body)
          .post()
          .go(callBack);
  },

  /**
   * Exchange a Refresh Token for an Access Token.
   * If you will be using the Refresh Token Grant, you will make a request to the Token endpoint to exchange the user’s refresh token for an access token.
   *
   * @param {string} refresh_token The refresh token that you would like to use to exchange for an access token.
   * @param {?string} client_id (Optional) The unique client identifier. The client Id is the Id of the FusionAuth Application in which you you are attempting to authenticate. This parameter is optional when the Authorization header is provided.
   * @param {?string} client_secret (Optional) The client secret. This value may optionally be provided in the request body instead of the Authorization header.
   * @param {?string} scope (Optional) This parameter is optional and if omitted, the same scope requested during the authorization request will be used. If provided the scopes must match those requested during the initial authorization request.
   * @param {?string} user_code (Optional) The end-user verification code. This code is required if using this endpoint to approve the Device Authorization.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  exchangeRefreshTokenForAccessToken: function(refresh_token, client_id, client_secret, scope, user_code, callBack) {
      var body = new FormData();
      body.append('refresh_token', refresh_token);
      body.append('client_id', client_id);
      body.append('client_secret', client_secret);
      body.append('grant_type', 'refresh_token');
      body.append('scope', scope);
      body.append('user_code', user_code);
      return this._startAnonymous()
          .uri('/oauth2/token')
          .setFormData(body)
          .post()
          .go(callBack);
  },

  /**
   * Exchange a refresh token for a new JWT.
   *
   * @param {RefreshRequest} request The refresh request.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  exchangeRefreshTokenForJWT: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/refresh')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Exchange User Credentials for a Token.
   * If you will be using the Resource Owner Password Credential Grant, you will make a request to the Token endpoint to exchange the user’s email and password for an access token.
   *
   * @param {string} username The login identifier of the user. The login identifier can be either the email or the username.
   * @param {string} password The user’s password.
   * @param {?string} client_id (Optional) The unique client identifier. The client Id is the Id of the FusionAuth Application in which you you are attempting to authenticate. This parameter is optional when the Authorization header is provided.
   * @param {?string} client_secret (Optional) The client secret. This value may optionally be provided in the request body instead of the Authorization header.
   * @param {?string} scope (Optional) This parameter is optional and if omitted, the same scope requested during the authorization request will be used. If provided the scopes must match those requested during the initial authorization request.
   * @param {?string} user_code (Optional) The end-user verification code. This code is required if using this endpoint to approve the Device Authorization.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  exchangeUserCredentialsForAccessToken: function(username, password, client_id, client_secret, scope, user_code, callBack) {
      var body = new FormData();
      body.append('username', username);
      body.append('password', password);
      body.append('client_id', client_id);
      body.append('client_secret', client_secret);
      body.append('grant_type', 'password');
      body.append('scope', scope);
      body.append('user_code', user_code);
      return this._startAnonymous()
          .uri('/oauth2/token')
          .setFormData(body)
          .post()
          .go(callBack);
  },

  /**
   * Begins the forgot password sequence, which kicks off an email to the user so that they can reset their password.
   *
   * @param {ForgotPasswordRequest} request The request that contains the information about the user so that they can be emailed.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  forgotPassword: function(request, callBack) {
      return this._start()
          .uri('/api/user/forgot-password')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Generate a new Email Verification Id to be used with the Verify Email API. This API will not attempt to send an
   * email to the User. This API may be used to collect the verificationId for use with a third party system.
   *
   * @param {string} email The email address of the user that needs a new verification email.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  generateEmailVerificationId: function(email, callBack) {
      return this._start()
          .uri('/api/user/verify-email')
          .urlParameter('email', email)
          .urlParameter('sendVerifyEmail', false)
          .put()
          .go(callBack);
  },

  /**
   * Generate a new RSA or EC key pair or an HMAC secret.
   *
   * @param {?UUIDString} keyId (Optional) The Id for the key. If not provided a secure random UUID will be generated.
   * @param {KeyRequest} request The request object that contains all of the information used to create the key.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  generateKey: function(keyId, request, callBack) {
      return this._start()
          .uri('/api/key/generate')
          .urlSegment(keyId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Generate a new Application Registration Verification Id to be used with the Verify Registration API. This API will not attempt to send an
   * email to the User. This API may be used to collect the verificationId for use with a third party system.
   *
   * @param {string} email The email address of the user that needs a new verification email.
   * @param {UUIDString} applicationId The Id of the application to be verified.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  generateRegistrationVerificationId: function(email, applicationId, callBack) {
      return this._start()
          .uri('/api/user/verify-registration')
          .urlParameter('email', email)
          .urlParameter('sendVerifyPasswordEmail', false)
          .urlParameter('applicationId', applicationId)
          .put()
          .go(callBack);
  },

  /**
   * Generate two-factor recovery codes for a user. Generating two-factor recovery codes will invalidate any existing recovery codes. 
   *
   * @param {UUIDString} userId The Id of the user to generate new Two Factor recovery codes.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  generateTwoFactorRecoveryCodes: function(userId, callBack) {
      return this._start()
          .header('Content-Type', 'text/plain')
          .uri('/api/user/two-factor/recovery-code')
          .urlSegment(userId)
          .post()
          .go(callBack);
  },

  /**
   * Generate a Two Factor secret that can be used to enable Two Factor authentication for a User. The response will contain
   * both the secret and a Base32 encoded form of the secret which can be shown to a User when using a 2 Step Authentication
   * application such as Google Authenticator.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  generateTwoFactorSecret: function(callBack) {
      return this._start()
          .uri('/api/two-factor/secret')
          .get()
          .go(callBack);
  },

  /**
   * Generate a Two Factor secret that can be used to enable Two Factor authentication for a User. The response will contain
   * both the secret and a Base32 encoded form of the secret which can be shown to a User when using a 2 Step Authentication
   * application such as Google Authenticator.
   *
   * @param {string} encodedJWT The encoded JWT (access token).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  generateTwoFactorSecretUsingJWT: function(encodedJWT, callBack) {
      return this._startAnonymous()
          .uri('/api/two-factor/secret')
          .authorization('Bearer ' + encodedJWT)
          .get()
          .go(callBack);
  },

  /**
   * Handles login via third-parties including Social login, external OAuth and OpenID Connect, and other
   * login systems.
   *
   * @param {IdentityProviderLoginRequest} request The third-party login request that contains information from the third-party login
   *    providers that FusionAuth uses to reconcile the user's account.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  identityProviderLogin: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/identity-provider/login')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Import an existing RSA or EC key pair or an HMAC secret.
   *
   * @param {?UUIDString} keyId (Optional) The Id for the key. If not provided a secure random UUID will be generated.
   * @param {KeyRequest} request The request object that contains all of the information used to create the key.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  importKey: function(keyId, request, callBack) {
      return this._start()
          .uri('/api/key/import')
          .urlSegment(keyId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Bulk imports refresh tokens. This request performs minimal validation and runs batch inserts of refresh tokens with the
   * expectation that each token represents a user that already exists and is registered for the corresponding FusionAuth
   * Application. This is done to increases the insert performance.
   * 
   * Therefore, if you encounter an error due to a database key violation, the response will likely offer a generic
   * explanation. If you encounter an error, you may optionally enable additional validation to receive a JSON response
   * body with specific validation errors. This will slow the request down but will allow you to identify the cause of
   * the failure. See the validateDbConstraints request parameter.
   *
   * @param {RefreshTokenImportRequest} request The request that contains all of the information about all of the refresh tokens to import.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  importRefreshTokens: function(request, callBack) {
      return this._start()
          .uri('/api/user/refresh-token/import')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Bulk imports users. This request performs minimal validation and runs batch inserts of users with the expectation
   * that each user does not yet exist and each registration corresponds to an existing FusionAuth Application. This is done to
   * increases the insert performance.
   * 
   * Therefore, if you encounter an error due to a database key violation, the response will likely offer
   * a generic explanation. If you encounter an error, you may optionally enable additional validation to receive a JSON response
   * body with specific validation errors. This will slow the request down but will allow you to identify the cause of the failure. See
   * the validateDbConstraints request parameter.
   *
   * @param {ImportRequest} request The request that contains all of the information about all of the users to import.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  importUsers: function(request, callBack) {
      return this._start()
          .uri('/api/user/import')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Inspect an access token issued by FusionAuth.
   *
   * @param {string} client_id The unique client identifier. The client Id is the Id of the FusionAuth Application for which this token was generated.
   * @param {string} token The access token returned by this OAuth provider as the result of a successful authentication.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  introspectAccessToken: function(client_id, token, callBack) {
      var body = new FormData();
      body.append('client_id', client_id);
      body.append('token', token);
      return this._startAnonymous()
          .uri('/oauth2/introspect')
          .setFormData(body)
          .post()
          .go(callBack);
  },

  /**
   * Issue a new access token (JWT) for the requested Application after ensuring the provided JWT is valid. A valid
   * access token is properly signed and not expired.
   * <p>
   * This API may be used in an SSO configuration to issue new tokens for another application after the user has
   * obtained a valid token from authentication.
   *
   * @param {UUIDString} applicationId The Application Id for which you are requesting a new access token be issued.
   * @param {string} encodedJWT The encoded JWT (access token).
   * @param {?string} refreshToken (Optional) An existing refresh token used to request a refresh token in addition to a JWT in the response.
   *    <p>The target application represented by the applicationId request parameter must have refresh
   *    tokens enabled in order to receive a refresh token in the response.</p>
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  issueJWT: function(applicationId, encodedJWT, refreshToken, callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/issue')
          .authorization('Bearer ' + encodedJWT)
          .urlParameter('applicationId', applicationId)
          .urlParameter('refreshToken', refreshToken)
          .get()
          .go(callBack);
  },

  /**
   * Authenticates a user to FusionAuth. 
   * 
   * This API optionally requires an API key. See <code>Application.loginConfiguration.requireAuthentication</code>.
   *
   * @param {LoginRequest} request The login request that contains the user credentials used to log them in.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  login: function(request, callBack) {
      return this._start()
          .uri('/api/login')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Sends a ping to FusionAuth indicating that the user was automatically logged into an application. When using
   * FusionAuth's SSO or your own, you should call this if the user is already logged in centrally, but accesses an
   * application where they no longer have a session. This helps correctly track login counts, times and helps with
   * reporting.
   *
   * @param {UUIDString} userId The Id of the user that was logged in.
   * @param {UUIDString} applicationId The Id of the application that they logged into.
   * @param {?string} callerIPAddress (Optional) The IP address of the end-user that is logging in. If a null value is provided
   *    the IP address will be that of the client or last proxy that sent the request.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  loginPing: function(userId, applicationId, callerIPAddress, callBack) {
      return this._start()
          .uri('/api/login')
          .urlSegment(userId)
          .urlSegment(applicationId)
          .urlParameter('ipAddress', callerIPAddress)
          .put()
          .go(callBack);
  },

  /**
   * The Logout API is intended to be used to remove the refresh token and access token cookies if they exist on the
   * client and revoke the refresh token stored. This API does nothing if the request does not contain an access
   * token or refresh token cookies.
   *
   * @param {boolean} global When this value is set to true all of the refresh tokens issued to the owner of the
   *    provided token will be revoked.
   * @param {?string} refreshToken (Optional) The refresh_token as a request parameter instead of coming in via a cookie.
   *    If provided this takes precedence over the cookie.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  logout: function(global, refreshToken, callBack) {
      return this._startAnonymous()
          .header('Content-Type', 'text/plain')
          .uri('/api/logout')
          .urlParameter('global', global)
          .urlParameter('refreshToken', refreshToken)
          .post()
          .go(callBack);
  },

  /**
   * Retrieves the identity provider for the given domain. A 200 response code indicates the domain is managed
   * by a registered identity provider. A 404 indicates the domain is not managed.
   *
   * @param {string} domain The domain or email address to lookup.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  lookupIdentityProvider: function(domain, callBack) {
      return this._start()
          .uri('/api/identity-provider/lookup')
          .urlParameter('domain', domain)
          .get()
          .go(callBack);
  },

  /**
   * Modifies a temporal user action by changing the expiration of the action and optionally adding a comment to the
   * action.
   *
   * @param {UUIDString} actionId The Id of the action to modify. This is technically the user action log id.
   * @param {ActionRequest} request The request that contains all of the information about the modification.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  modifyAction: function(actionId, request, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlSegment(actionId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Complete a login request using a passwordless code
   *
   * @param {PasswordlessLoginRequest} request The passwordless login request that contains all of the information used to complete login.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  passwordlessLogin: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/passwordless/login')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Updates an authentication API key by given id
   *
   * @param {UUIDString} keyId The Id of the authentication key. If not provided a secure random api key will be generated.
   * @param {APIKeyRequest} request The request object that contains all of the information needed to create the APIKey.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchAPIKey: function(keyId, request, callBack) {
      return this._start()
          .uri('/api/api-key')
          .urlSegment(keyId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the application with the given Id.
   *
   * @param {UUIDString} applicationId The Id of the application to update.
   * @param {ApplicationRequest} request The request that contains just the new application information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchApplication: function(applicationId, request, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the application role with the given id for the application.
   *
   * @param {UUIDString} applicationId The Id of the application that the role belongs to.
   * @param {UUIDString} roleId The Id of the role to update.
   * @param {ApplicationRequest} request The request that contains just the new role information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchApplicationRole: function(applicationId, roleId, request, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment("role")
          .urlSegment(roleId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the connector with the given Id.
   *
   * @param {UUIDString} connectorId The Id of the connector to update.
   * @param {ConnectorRequest} request The request that contains just the new connector information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchConnector: function(connectorId, request, callBack) {
      return this._start()
          .uri('/api/connector')
          .urlSegment(connectorId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the consent with the given Id.
   *
   * @param {UUIDString} consentId The Id of the consent to update.
   * @param {ConsentRequest} request The request that contains just the new consent information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchConsent: function(consentId, request, callBack) {
      return this._start()
          .uri('/api/consent')
          .urlSegment(consentId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the email template with the given Id.
   *
   * @param {UUIDString} emailTemplateId The Id of the email template to update.
   * @param {EmailTemplateRequest} request The request that contains just the new email template information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchEmailTemplate: function(emailTemplateId, request, callBack) {
      return this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the Entity Type with the given Id.
   *
   * @param {UUIDString} entityTypeId The Id of the Entity Type to update.
   * @param {EntityTypeRequest} request The request that contains just the new Entity Type information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchEntityType: function(entityTypeId, request, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the group with the given Id.
   *
   * @param {UUIDString} groupId The Id of the group to update.
   * @param {GroupRequest} request The request that contains just the new group information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchGroup: function(groupId, request, callBack) {
      return this._start()
          .uri('/api/group')
          .urlSegment(groupId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the identity provider with the given Id.
   *
   * @param {UUIDString} identityProviderId The Id of the identity provider to update.
   * @param {IdentityProviderRequest} request The request object that contains just the updated identity provider information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchIdentityProvider: function(identityProviderId, request, callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .urlSegment(identityProviderId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the available integrations.
   *
   * @param {IntegrationRequest} request The request that contains just the new integration information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchIntegrations: function(request, callBack) {
      return this._start()
          .uri('/api/integration')
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the lambda with the given Id.
   *
   * @param {UUIDString} lambdaId The Id of the lambda to update.
   * @param {LambdaRequest} request The request that contains just the new lambda information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchLambda: function(lambdaId, request, callBack) {
      return this._start()
          .uri('/api/lambda')
          .urlSegment(lambdaId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the message template with the given Id.
   *
   * @param {UUIDString} messageTemplateId The Id of the message template to update.
   * @param {MessageTemplateRequest} request The request that contains just the new message template information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchMessageTemplate: function(messageTemplateId, request, callBack) {
      return this._start()
          .uri('/api/message/template')
          .urlSegment(messageTemplateId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the messenger with the given Id.
   *
   * @param {UUIDString} messengerId The Id of the messenger to update.
   * @param {MessengerRequest} request The request that contains just the new messenger information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchMessenger: function(messengerId, request, callBack) {
      return this._start()
          .uri('/api/messenger')
          .urlSegment(messengerId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the registration for the user with the given id and the application defined in the request.
   *
   * @param {UUIDString} userId The Id of the user whose registration is going to be updated.
   * @param {RegistrationRequest} request The request that contains just the new registration information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchRegistration: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the system configuration.
   *
   * @param {SystemConfigurationRequest} request The request that contains just the new system configuration information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchSystemConfiguration: function(request, callBack) {
      return this._start()
          .uri('/api/system-configuration')
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the tenant with the given Id.
   *
   * @param {UUIDString} tenantId The Id of the tenant to update.
   * @param {TenantRequest} request The request that contains just the new tenant information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchTenant: function(tenantId, request, callBack) {
      return this._start()
          .uri('/api/tenant')
          .urlSegment(tenantId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the theme with the given Id.
   *
   * @param {UUIDString} themeId The Id of the theme to update.
   * @param {ThemeRequest} request The request that contains just the new theme information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchTheme: function(themeId, request, callBack) {
      return this._start()
          .uri('/api/theme')
          .urlSegment(themeId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the user with the given Id.
   *
   * @param {UUIDString} userId The Id of the user to update.
   * @param {UserRequest} request The request that contains just the new user information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchUser: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the user action with the given Id.
   *
   * @param {UUIDString} userActionId The Id of the user action to update.
   * @param {UserActionRequest} request The request that contains just the new user action information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchUserAction: function(userActionId, request, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, the user action reason with the given Id.
   *
   * @param {UUIDString} userActionReasonId The Id of the user action reason to update.
   * @param {UserActionReasonRequest} request The request that contains just the new user action reason information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchUserActionReason: function(userActionReasonId, request, callBack) {
      return this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Updates, via PATCH, a single User consent by Id.
   *
   * @param {UUIDString} userConsentId The User Consent Id
   * @param {UserConsentRequest} request The request that contains just the new user consent information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  patchUserConsent: function(userConsentId, request, callBack) {
      return this._start()
          .uri('/api/user/consent')
          .urlSegment(userConsentId)
          .setJSONBody(request)
          .patch()
          .go(callBack);
  },

  /**
   * Reactivates the application with the given Id.
   *
   * @param {UUIDString} applicationId The Id of the application to reactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  reactivateApplication: function(applicationId, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlParameter('reactivate', true)
          .put()
          .go(callBack);
  },

  /**
   * Reactivates the user with the given Id.
   *
   * @param {UUIDString} userId The Id of the user to reactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  reactivateUser: function(userId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .urlParameter('reactivate', true)
          .put()
          .go(callBack);
  },

  /**
   * Reactivates the user action with the given Id.
   *
   * @param {UUIDString} userActionId The Id of the user action to reactivate.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  reactivateUserAction: function(userActionId, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .urlParameter('reactivate', true)
          .put()
          .go(callBack);
  },

  /**
   * Reconcile a User to FusionAuth using JWT issued from another Identity Provider.
   *
   * @param {IdentityProviderLoginRequest} request The reconcile request that contains the data to reconcile the User.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  reconcileJWT: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/reconcile')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Request a refresh of the Entity search index. This API is not generally necessary and the search index will become consistent in a
   * reasonable amount of time. There may be scenarios where you may wish to manually request an index refresh. One example may be 
   * if you are using the Search API or Delete Tenant API immediately following a Entity Create etc, you may wish to request a refresh to
   *  ensure the index immediately current before making a query request to the search index.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  refreshEntitySearchIndex: function(callBack) {
      return this._start()
          .uri('/api/entity/search')
          .put()
          .go(callBack);
  },

  /**
   * Request a refresh of the User search index. This API is not generally necessary and the search index will become consistent in a
   * reasonable amount of time. There may be scenarios where you may wish to manually request an index refresh. One example may be 
   * if you are using the Search API or Delete Tenant API immediately following a User Create etc, you may wish to request a refresh to
   *  ensure the index immediately current before making a query request to the search index.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  refreshUserSearchIndex: function(callBack) {
      return this._start()
          .uri('/api/user/search')
          .put()
          .go(callBack);
  },

  /**
   * Regenerates any keys that are used by the FusionAuth Reactor.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  regenerateReactorKeys: function(callBack) {
      return this._start()
          .uri('/api/reactor')
          .put()
          .go(callBack);
  },

  /**
   * Registers a user for an application. If you provide the User and the UserRegistration object on this request, it
   * will create the user as well as register them for the application. This is called a Full Registration. However, if
   * you only provide the UserRegistration object, then the user must already exist and they will be registered for the
   * application. The user id can also be provided and it will either be used to look up an existing user or it will be
   * used for the newly created User.
   *
   * @param {?UUIDString} userId (Optional) The Id of the user being registered for the application and optionally created.
   * @param {RegistrationRequest} request The request that optionally contains the User and must contain the UserRegistration.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  register: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Requests Elasticsearch to delete and rebuild the index for FusionAuth users or entities. Be very careful when running this request as it will 
   * increase the CPU and I/O load on your database until the operation completes. Generally speaking you do not ever need to run this operation unless 
   * instructed by FusionAuth support, or if you are migrating a database another system and you are not brining along the Elasticsearch index. 
   * 
   * You have been warned.
   *
   * @param {ReindexRequest} request The request that contains the index name.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  reindex: function(request, callBack) {
      return this._start()
          .uri('/api/system/reindex')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Removes a user from the family with the given id.
   *
   * @param {UUIDString} familyId The id of the family to remove the user from.
   * @param {UUIDString} userId The id of the user to remove from the family.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  removeUserFromFamily: function(familyId, userId, callBack) {
      return this._start()
          .uri('/api/user/family')
          .urlSegment(familyId)
          .urlSegment(userId)
          .delete()
          .go(callBack);
  },

  /**
   * Re-sends the verification email to the user.
   *
   * @param {string} email The email address of the user that needs a new verification email.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  resendEmailVerification: function(email, callBack) {
      return this._start()
          .uri('/api/user/verify-email')
          .urlParameter('email', email)
          .put()
          .go(callBack);
  },

  /**
   * Re-sends the verification email to the user. If the Application has configured a specific email template this will be used
   * instead of the tenant configuration.
   *
   * @param {UUIDString} applicationId The unique Application Id to used to resolve an application specific email template.
   * @param {string} email The email address of the user that needs a new verification email.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  resendEmailVerificationWithApplicationTemplate: function(applicationId, email, callBack) {
      return this._start()
          .uri('/api/user/verify-email')
          .urlParameter('applicationId', applicationId)
          .urlParameter('email', email)
          .put()
          .go(callBack);
  },

  /**
   * Re-sends the application registration verification email to the user.
   *
   * @param {string} email The email address of the user that needs a new verification email.
   * @param {UUIDString} applicationId The Id of the application to be verified.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  resendRegistrationVerification: function(email, applicationId, callBack) {
      return this._start()
          .uri('/api/user/verify-registration')
          .urlParameter('email', email)
          .urlParameter('applicationId', applicationId)
          .put()
          .go(callBack);
  },

  /**
   * Retrieves an authentication API key for the given id
   *
   * @param {UUIDString} keyId The Id of the API key to retrieve.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveAPIKey: function(keyId, callBack) {
      return this._start()
          .uri('/api/api-key')
          .urlSegment(keyId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves a single action log (the log of a user action that was taken on a user previously) for the given Id.
   *
   * @param {UUIDString} actionId The Id of the action to retrieve.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveAction: function(actionId, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlSegment(actionId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the actions for the user with the given Id. This will return all time based actions that are active,
   * and inactive as well as non-time based actions.
   *
   * @param {UUIDString} userId The Id of the user to fetch the actions for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveActions: function(userId, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the actions for the user with the given Id that are currently preventing the User from logging in.
   *
   * @param {UUIDString} userId The Id of the user to fetch the actions for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveActionsPreventingLogin: function(userId, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlParameter('userId', userId)
          .urlParameter('preventingLogin', true)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the actions for the user with the given Id that are currently active.
   * An active action means one that is time based and has not been canceled, and has not ended.
   *
   * @param {UUIDString} userId The Id of the user to fetch the actions for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveActiveActions: function(userId, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlParameter('userId', userId)
          .urlParameter('active', true)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the application for the given id or all of the applications if the id is null.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveApplication: function(applicationId, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the applications.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveApplications: function(callBack) {
      return this._start()
          .uri('/api/application')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves a single audit log for the given Id.
   *
   * @param {number} auditLogId The Id of the audit log to retrieve.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveAuditLog: function(auditLogId, callBack) {
      return this._start()
          .uri('/api/system/audit-log')
          .urlSegment(auditLogId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the connector with the given Id.
   *
   * @param {UUIDString} connectorId The Id of the connector.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveConnector: function(connectorId, callBack) {
      return this._start()
          .uri('/api/connector')
          .urlSegment(connectorId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the connectors.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveConnectors: function(callBack) {
      return this._start()
          .uri('/api/connector')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the Consent for the given Id.
   *
   * @param {UUIDString} consentId The Id of the consent.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveConsent: function(consentId, callBack) {
      return this._start()
          .uri('/api/consent')
          .urlSegment(consentId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the consent.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveConsents: function(callBack) {
      return this._start()
          .uri('/api/consent')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the daily active user report between the two instants. If you specify an application id, it will only
   * return the daily active counts for that application.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {number} start The start instant as UTC milliseconds since Epoch.
   * @param {number} end The end instant as UTC milliseconds since Epoch.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveDailyActiveReport: function(applicationId, start, end, callBack) {
      return this._start()
          .uri('/api/report/daily-active-user')
          .urlParameter('applicationId', applicationId)
          .urlParameter('start', start)
          .urlParameter('end', end)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the email template for the given Id. If you don't specify the id, this will return all of the email templates.
   *
   * @param {?UUIDString} emailTemplateId (Optional) The Id of the email template.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEmailTemplate: function(emailTemplateId, callBack) {
      return this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .get()
          .go(callBack);
  },

  /**
   * Creates a preview of the email template provided in the request. This allows you to preview an email template that
   * hasn't been saved to the database yet. The entire email template does not need to be provided on the request. This
   * will create the preview based on whatever is given.
   *
   * @param {PreviewRequest} request The request that contains the email template and optionally a locale to render it in.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEmailTemplatePreview: function(request, callBack) {
      return this._start()
          .uri('/api/email/template/preview')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Retrieves all of the email templates.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEmailTemplates: function(callBack) {
      return this._start()
          .uri('/api/email/template')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the Entity for the given Id.
   *
   * @param {UUIDString} entityId The Id of the Entity.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEntity: function(entityId, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves an Entity Grant for the given Entity and User/Entity.
   *
   * @param {UUIDString} entityId The Id of the Entity.
   * @param {?UUIDString} recipientEntityId (Optional) The Id of the Entity that the Entity Grant is for.
   * @param {?UUIDString} userId (Optional) The Id of the User that the Entity Grant is for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEntityGrant: function(entityId, recipientEntityId, userId, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .urlSegment("grant")
          .urlParameter('recipientEntityId', recipientEntityId)
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the Entity Type for the given Id.
   *
   * @param {UUIDString} entityTypeId The Id of the Entity Type.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEntityType: function(entityTypeId, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the Entity Types.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEntityTypes: function(callBack) {
      return this._start()
          .uri('/api/entity/type')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves a single event log for the given Id.
   *
   * @param {number} eventLogId The Id of the event log to retrieve.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveEventLog: function(eventLogId, callBack) {
      return this._start()
          .uri('/api/system/event-log')
          .urlSegment(eventLogId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the families that a user belongs to.
   *
   * @param {UUIDString} userId The User's id
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveFamilies: function(userId, callBack) {
      return this._start()
          .uri('/api/user/family')
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the members of a family by the unique Family Id.
   *
   * @param {UUIDString} familyId The unique Id of the Family.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveFamilyMembersByFamilyId: function(familyId, callBack) {
      return this._start()
          .uri('/api/user/family')
          .urlSegment(familyId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the form with the given Id.
   *
   * @param {UUIDString} formId The Id of the form.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveForm: function(formId, callBack) {
      return this._start()
          .uri('/api/form')
          .urlSegment(formId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the form field with the given Id.
   *
   * @param {UUIDString} fieldId The Id of the form field.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveFormField: function(fieldId, callBack) {
      return this._start()
          .uri('/api/form/field')
          .urlSegment(fieldId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the forms fields
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveFormFields: function(callBack) {
      return this._start()
          .uri('/api/form/field')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the forms.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveForms: function(callBack) {
      return this._start()
          .uri('/api/form')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the group for the given Id.
   *
   * @param {UUIDString} groupId The Id of the group.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveGroup: function(groupId, callBack) {
      return this._start()
          .uri('/api/group')
          .urlSegment(groupId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the groups.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveGroups: function(callBack) {
      return this._start()
          .uri('/api/group')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the identity provider for the given id or all of the identity providers if the id is null.
   *
   * @param {UUIDString} identityProviderId The identity provider Id.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveIdentityProvider: function(identityProviderId, callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .urlSegment(identityProviderId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves one or more identity provider for the given type. For types such as Google, Facebook, Twitter and LinkedIn, only a single 
   * identity provider can exist. For types such as OpenID Connect and SAMLv2 more than one identity provider can be configured so this request 
   * may return multiple identity providers.
   *
   * @param {IdentityProviderType} type The type of the identity provider.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveIdentityProviderByType: function(type, callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .urlParameter('type', type)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the identity providers.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveIdentityProviders: function(callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the actions for the user with the given Id that are currently inactive.
   * An inactive action means one that is time based and has been canceled or has expired, or is not time based.
   *
   * @param {UUIDString} userId The Id of the user to fetch the actions for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveInactiveActions: function(userId, callBack) {
      return this._start()
          .uri('/api/user/action')
          .urlParameter('userId', userId)
          .urlParameter('active', false)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the applications that are currently inactive.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveInactiveApplications: function(callBack) {
      return this._start()
          .uri('/api/application')
          .urlParameter('inactive', true)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the user actions that are currently inactive.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveInactiveUserActions: function(callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlParameter('inactive', true)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the available integrations.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveIntegration: function(callBack) {
      return this._start()
          .uri('/api/integration')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the Public Key configured for verifying JSON Web Tokens (JWT) by the key Id (kid).
   *
   * @param {string} keyId The Id of the public key (kid).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveJWTPublicKey: function(keyId, callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/public-key')
          .urlParameter('kid', keyId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the Public Key configured for verifying the JSON Web Tokens (JWT) issued by the Login API by the Application Id.
   *
   * @param {string} applicationId The Id of the Application for which this key is used.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveJWTPublicKeyByApplicationId: function(applicationId, callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/public-key')
          .urlParameter('applicationId', applicationId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all Public Keys configured for verifying JSON Web Tokens (JWT).
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveJWTPublicKeys: function(callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/public-key')
          .get()
          .go(callBack);
  },

  /**
   * Returns public keys used by FusionAuth to cryptographically verify JWTs using the JSON Web Key format.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveJsonWebKeySet: function(callBack) {
      return this._startAnonymous()
          .uri('/.well-known/jwks.json')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the key for the given Id.
   *
   * @param {UUIDString} keyId The Id of the key.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveKey: function(keyId, callBack) {
      return this._start()
          .uri('/api/key')
          .urlSegment(keyId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the keys.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveKeys: function(callBack) {
      return this._start()
          .uri('/api/key')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the lambda for the given Id.
   *
   * @param {UUIDString} lambdaId The Id of the lambda.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveLambda: function(lambdaId, callBack) {
      return this._start()
          .uri('/api/lambda')
          .urlSegment(lambdaId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the lambdas.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveLambdas: function(callBack) {
      return this._start()
          .uri('/api/lambda')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the lambdas for the provided type.
   *
   * @param {LambdaType} type The type of the lambda to return.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveLambdasByType: function(type, callBack) {
      return this._start()
          .uri('/api/lambda')
          .urlParameter('type', type)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the login report between the two instants. If you specify an application id, it will only return the
   * login counts for that application.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {number} start The start instant as UTC milliseconds since Epoch.
   * @param {number} end The end instant as UTC milliseconds since Epoch.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveLoginReport: function(applicationId, start, end, callBack) {
      return this._start()
          .uri('/api/report/login')
          .urlParameter('applicationId', applicationId)
          .urlParameter('start', start)
          .urlParameter('end', end)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the message template for the given Id. If you don't specify the id, this will return all of the message templates.
   *
   * @param {?UUIDString} messageTemplateId (Optional) The Id of the message template.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveMessageTemplate: function(messageTemplateId, callBack) {
      return this._start()
          .uri('/api/message/template')
          .urlSegment(messageTemplateId)
          .get()
          .go(callBack);
  },

  /**
   * Creates a preview of the message template provided in the request, normalized to a given locale.
   *
   * @param {PreviewMessageTemplateRequest} request The request that contains the email template and optionally a locale to render it in.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveMessageTemplatePreview: function(request, callBack) {
      return this._start()
          .uri('/api/message/template/preview')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Retrieves all of the message templates.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveMessageTemplates: function(callBack) {
      return this._start()
          .uri('/api/message/template')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the messenger with the given Id.
   *
   * @param {UUIDString} messengerId The Id of the messenger.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveMessenger: function(messengerId, callBack) {
      return this._start()
          .uri('/api/messenger')
          .urlSegment(messengerId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the messengers.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveMessengers: function(callBack) {
      return this._start()
          .uri('/api/messenger')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the monthly active user report between the two instants. If you specify an application id, it will only
   * return the monthly active counts for that application.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {number} start The start instant as UTC milliseconds since Epoch.
   * @param {number} end The end instant as UTC milliseconds since Epoch.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveMonthlyActiveReport: function(applicationId, start, end, callBack) {
      return this._start()
          .uri('/api/report/monthly-active-user')
          .urlParameter('applicationId', applicationId)
          .urlParameter('start', start)
          .urlParameter('end', end)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the Oauth2 configuration for the application for the given Application Id.
   *
   * @param {UUIDString} applicationId The Id of the Application to retrieve OAuth configuration.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveOauthConfiguration: function(applicationId, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment("oauth-configuration")
          .get()
          .go(callBack);
  },

  /**
   * Returns the well known OpenID Configuration JSON document
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveOpenIdConfiguration: function(callBack) {
      return this._startAnonymous()
          .uri('/.well-known/openid-configuration')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the password validation rules for a specific tenant. This method requires a tenantId to be provided 
   * through the use of a Tenant scoped API key or an HTTP header X-FusionAuth-TenantId to specify the Tenant Id.
   * 
   * This API does not require an API key.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrievePasswordValidationRules: function(callBack) {
      return this._startAnonymous()
          .uri('/api/tenant/password-validation-rules')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the password validation rules for a specific tenant.
   * 
   * This API does not require an API key.
   *
   * @param {UUIDString} tenantId The Id of the tenant.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrievePasswordValidationRulesWithTenantId: function(tenantId, callBack) {
      return this._startAnonymous()
          .uri('/api/tenant/password-validation-rules')
          .urlSegment(tenantId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the children for the given parent email address.
   *
   * @param {string} parentEmail The email of the parent.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrievePendingChildren: function(parentEmail, callBack) {
      return this._start()
          .uri('/api/user/family/pending')
          .urlParameter('parentEmail', parentEmail)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the FusionAuth Reactor status.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveReactorStatus: function(callBack) {
      return this._start()
          .uri('/api/reactor')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the last number of login records.
   *
   * @param {number} offset The initial record. e.g. 0 is the last login, 100 will be the 100th most recent login.
   * @param {number} limit (Optional, defaults to 10) The number of records to retrieve.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveRecentLogins: function(offset, limit, callBack) {
      return this._start()
          .uri('/api/user/recent-login')
          .urlParameter('offset', offset)
          .urlParameter('limit', limit)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves a single refresh token by unique Id. This is not the same thing as the string value of the refresh token, if you have that, you already have what you need..
   *
   * @param {UUIDString} userId The Id of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveRefreshTokenById: function(userId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlSegment(userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the refresh tokens that belong to the user with the given Id.
   *
   * @param {UUIDString} userId The Id of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveRefreshTokens: function(userId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user registration for the user with the given id and the given application id.
   *
   * @param {UUIDString} userId The Id of the user.
   * @param {UUIDString} applicationId The Id of the application.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveRegistration: function(userId, applicationId, callBack) {
      return this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .urlSegment(applicationId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the registration report between the two instants. If you specify an application id, it will only return
   * the registration counts for that application.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {number} start The start instant as UTC milliseconds since Epoch.
   * @param {number} end The end instant as UTC milliseconds since Epoch.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveRegistrationReport: function(applicationId, start, end, callBack) {
      return this._start()
          .uri('/api/report/registration')
          .urlParameter('applicationId', applicationId)
          .urlParameter('start', start)
          .urlParameter('end', end)
          .get()
          .go(callBack);
  },

  /**
   * Retrieve the status of a re-index process. A status code of 200 indicates the re-index is in progress, a status code of  
   * 404 indicates no re-index is in progress.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveReindexStatus: function(callBack) {
      return this._start()
          .uri('/api/system/reindex')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the system configuration.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveSystemConfiguration: function(callBack) {
      return this._start()
          .uri('/api/system-configuration')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the tenant for the given Id.
   *
   * @param {UUIDString} tenantId The Id of the tenant.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveTenant: function(tenantId, callBack) {
      return this._start()
          .uri('/api/tenant')
          .urlSegment(tenantId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the tenants.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveTenants: function(callBack) {
      return this._start()
          .uri('/api/tenant')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the theme for the given Id.
   *
   * @param {UUIDString} themeId The Id of the theme.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveTheme: function(themeId, callBack) {
      return this._start()
          .uri('/api/theme')
          .urlSegment(themeId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the themes.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveThemes: function(callBack) {
      return this._start()
          .uri('/api/theme')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the totals report. This contains all of the total counts for each application and the global registration
   * count.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveTotalReport: function(callBack) {
      return this._start()
          .uri('/api/report/totals')
          .get()
          .go(callBack);
  },

  /**
   * Retrieve two-factor recovery codes for a user.
   *
   * @param {UUIDString} userId The Id of the user to retrieve Two Factor recovery codes.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveTwoFactorRecoveryCodes: function(userId, callBack) {
      return this._start()
          .uri('/api/user/two-factor/recovery-code')
          .urlSegment(userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user for the given Id.
   *
   * @param {UUIDString} userId The Id of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUser: function(userId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user action for the given Id. If you pass in null for the id, this will return all of the user
   * actions.
   *
   * @param {?UUIDString} userActionId (Optional) The Id of the user action.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserAction: function(userActionId, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user action reason for the given Id. If you pass in null for the id, this will return all of the user
   * action reasons.
   *
   * @param {?UUIDString} userActionReasonId (Optional) The Id of the user action reason.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserActionReason: function(userActionReasonId, callBack) {
      return this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all the user action reasons.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserActionReasons: function(callBack) {
      return this._start()
          .uri('/api/user-action-reason')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the user actions.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserActions: function(callBack) {
      return this._start()
          .uri('/api/user-action')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user by a change password Id. The intended use of this API is to retrieve a user after the forgot
   * password workflow has been initiated and you may not know the user's email or username.
   *
   * @param {string} changePasswordId The unique change password Id that was sent via email or returned by the Forgot Password API.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserByChangePasswordId: function(changePasswordId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlParameter('changePasswordId', changePasswordId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user for the given email.
   *
   * @param {string} email The email of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserByEmail: function(email, callBack) {
      return this._start()
          .uri('/api/user')
          .urlParameter('email', email)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user for the loginId. The loginId can be either the username or the email.
   *
   * @param {string} loginId The email or username of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserByLoginId: function(loginId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlParameter('loginId', loginId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user for the given username.
   *
   * @param {string} username The username of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserByUsername: function(username, callBack) {
      return this._start()
          .uri('/api/user')
          .urlParameter('username', username)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user by a verificationId. The intended use of this API is to retrieve a user after the forgot
   * password workflow has been initiated and you may not know the user's email or username.
   *
   * @param {string} verificationId The unique verification Id that has been set on the user object.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserByVerificationId: function(verificationId, callBack) {
      return this._start()
          .uri('/api/user')
          .urlParameter('verificationId', verificationId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the comments for the user with the given Id.
   *
   * @param {UUIDString} userId The Id of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserComments: function(userId, callBack) {
      return this._start()
          .uri('/api/user/comment')
          .urlSegment(userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieve a single User consent by Id.
   *
   * @param {UUIDString} userConsentId The User consent Id
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserConsent: function(userConsentId, callBack) {
      return this._start()
          .uri('/api/user/consent')
          .urlSegment(userConsentId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all of the consents for a User.
   *
   * @param {UUIDString} userId The User's Id
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserConsents: function(userId, callBack) {
      return this._start()
          .uri('/api/user/consent')
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Call the UserInfo endpoint to retrieve User Claims from the access token issued by FusionAuth.
   *
   * @param {string} encodedJWT The encoded JWT (access token).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserInfoFromAccessToken: function(encodedJWT, callBack) {
      return this._startAnonymous()
          .uri('/oauth2/userinfo')
          .authorization('Bearer ' + encodedJWT)
          .get()
          .go(callBack);
  },

  /**
   * Retrieve a single Identity Provider user (link).
   *
   * @param {UUIDString} identityProviderId The unique Id of the identity provider.
   * @param {string} identityProviderUserId The unique Id of the user in the 3rd party identity provider.
   * @param {UUIDString} userId The unique Id of the FusionAuth user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserLink: function(identityProviderId, identityProviderUserId, userId, callBack) {
      return this._start()
          .uri('/api/identity-provider/link')
          .urlParameter('identityProviderId', identityProviderId)
          .urlParameter('identityProviderUserId', identityProviderUserId)
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieve all Identity Provider users (links) for the user. Specify the optional identityProviderId to retrieve links for a particular IdP.
   *
   * @param {?UUIDString} identityProviderId (Optional) The unique Id of the identity provider. Specify this value to reduce the links returned to those for a particular IdP.
   * @param {UUIDString} userId The unique Id of the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserLinksByUserId: function(identityProviderId, userId, callBack) {
      return this._start()
          .uri('/api/identity-provider/link')
          .urlParameter('identityProviderId', identityProviderId)
          .urlParameter('userId', userId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the login report between the two instants for a particular user by Id. If you specify an application id, it will only return the
   * login counts for that application.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {UUIDString} userId The userId id.
   * @param {number} start The start instant as UTC milliseconds since Epoch.
   * @param {number} end The end instant as UTC milliseconds since Epoch.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserLoginReport: function(applicationId, userId, start, end, callBack) {
      return this._start()
          .uri('/api/report/login')
          .urlParameter('applicationId', applicationId)
          .urlParameter('userId', userId)
          .urlParameter('start', start)
          .urlParameter('end', end)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the login report between the two instants for a particular user by login Id. If you specify an application id, it will only return the
   * login counts for that application.
   *
   * @param {?UUIDString} applicationId (Optional) The application id.
   * @param {string} loginId The userId id.
   * @param {number} start The start instant as UTC milliseconds since Epoch.
   * @param {number} end The end instant as UTC milliseconds since Epoch.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserLoginReportByLoginId: function(applicationId, loginId, start, end, callBack) {
      return this._start()
          .uri('/api/report/login')
          .urlParameter('applicationId', applicationId)
          .urlParameter('loginId', loginId)
          .urlParameter('start', start)
          .urlParameter('end', end)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the last number of login records for a user.
   *
   * @param {UUIDString} userId The Id of the user.
   * @param {number} offset The initial record. e.g. 0 is the last login, 100 will be the 100th most recent login.
   * @param {number} limit (Optional, defaults to 10) The number of records to retrieve.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserRecentLogins: function(userId, offset, limit, callBack) {
      return this._start()
          .uri('/api/user/recent-login')
          .urlParameter('userId', userId)
          .urlParameter('offset', offset)
          .urlParameter('limit', limit)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the user for the given Id. This method does not use an API key, instead it uses a JSON Web Token (JWT) for authentication.
   *
   * @param {string} encodedJWT The encoded JWT (access token).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveUserUsingJWT: function(encodedJWT, callBack) {
      return this._startAnonymous()
          .uri('/api/user')
          .authorization('Bearer ' + encodedJWT)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the FusionAuth version string.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveVersion: function(callBack) {
      return this._start()
          .uri('/api/system/version')
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the webhook for the given Id. If you pass in null for the id, this will return all the webhooks.
   *
   * @param {?UUIDString} webhookId (Optional) The Id of the webhook.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveWebhook: function(webhookId, callBack) {
      return this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves all the webhooks.
   *
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  retrieveWebhooks: function(callBack) {
      return this._start()
          .uri('/api/webhook')
          .get()
          .go(callBack);
  },

  /**
   * Revokes refresh tokens.
   * 
   * Usage examples:
   *   - Delete a single refresh token, pass in only the token.
   *       revokeRefreshToken(token)
   * 
   *   - Delete all refresh tokens for a user, pass in only the userId.
   *       revokeRefreshToken(null, userId)
   * 
   *   - Delete all refresh tokens for a user for a specific application, pass in both the userId and the applicationId.
   *       revokeRefreshToken(null, userId, applicationId)
   * 
   *   - Delete all refresh tokens for an application
   *       revokeRefreshToken(null, null, applicationId)
   * 
   * Note: <code>null</code> may be handled differently depending upon the programming language.
   * 
   * See also: (method names may vary by language... but you'll figure it out)
   * 
   *  - revokeRefreshTokenById
   *  - revokeRefreshTokenByToken
   *  - revokeRefreshTokensByUserId
   *  - revokeRefreshTokensByApplicationId
   *  - revokeRefreshTokensByUserIdForApplication
   *
   * @param {?string} token (Optional) The refresh token to delete.
   * @param {?UUIDString} userId (Optional) The user id whose tokens to delete.
   * @param {?UUIDString} applicationId (Optional) The application id of the tokens to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeRefreshToken: function(token, userId, applicationId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('token', token)
          .urlParameter('userId', userId)
          .urlParameter('applicationId', applicationId)
          .delete()
          .go(callBack);
  },

  /**
   * Revokes a single refresh token by the unique Id. The unique Id is not sensitive as it cannot be used to obtain another JWT.
   *
   * @param {UUIDString} tokenId The unique Id of the token to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeRefreshTokenById: function(tokenId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlSegment(tokenId)
          .delete()
          .go(callBack);
  },

  /**
   * Revokes a single refresh token by using the actual refresh token value. This refresh token value is sensitive, so  be careful with this API request.
   *
   * @param {string} token The refresh token to delete.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeRefreshTokenByToken: function(token, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('token', token)
          .delete()
          .go(callBack);
  },

  /**
   * Revoke all refresh tokens that belong to an application by applicationId.
   *
   * @param {UUIDString} applicationId The unique Id of the application that you want to delete all refresh tokens for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeRefreshTokensByApplicationId: function(applicationId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('applicationId', applicationId)
          .delete()
          .go(callBack);
  },

  /**
   * Revoke all refresh tokens that belong to a user by user Id.
   *
   * @param {UUIDString} userId The unique Id of the user that you want to delete all refresh tokens for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeRefreshTokensByUserId: function(userId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('userId', userId)
          .delete()
          .go(callBack);
  },

  /**
   * Revoke all refresh tokens that belong to a user by user Id for a specific application by applicationId.
   *
   * @param {UUIDString} userId The unique Id of the user that you want to delete all refresh tokens for.
   * @param {UUIDString} applicationId The unique Id of the application that you want to delete refresh tokens for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeRefreshTokensByUserIdForApplication: function(userId, applicationId, callBack) {
      return this._start()
          .uri('/api/jwt/refresh')
          .urlParameter('userId', userId)
          .urlParameter('applicationId', applicationId)
          .delete()
          .go(callBack);
  },

  /**
   * Revokes a single User consent by Id.
   *
   * @param {UUIDString} userConsentId The User Consent Id
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  revokeUserConsent: function(userConsentId, callBack) {
      return this._start()
          .uri('/api/user/consent')
          .urlSegment(userConsentId)
          .delete()
          .go(callBack);
  },

  /**
   * Searches the audit logs with the specified criteria and pagination.
   *
   * @param {AuditLogSearchRequest} request The search criteria and pagination information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchAuditLogs: function(request, callBack) {
      return this._start()
          .uri('/api/system/audit-log/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Searches entities with the specified criteria and pagination.
   *
   * @param {EntitySearchRequest} request The search criteria and pagination information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchEntities: function(request, callBack) {
      return this._start()
          .uri('/api/entity/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Retrieves the entities for the given ids. If any id is invalid, it is ignored.
   *
   * @param {Array<string>} ids The entity ids to search for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchEntitiesByIds: function(ids, callBack) {
      return this._start()
          .uri('/api/entity/search')
          .urlParameter('ids', ids)
          .get()
          .go(callBack);
  },

  /**
   * Searches Entity Grants with the specified criteria and pagination.
   *
   * @param {EntityGrantSearchRequest} request The search criteria and pagination information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchEntityGrants: function(request, callBack) {
      return this._start()
          .uri('/api/entity/grant/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Searches the entity types with the specified criteria and pagination.
   *
   * @param {EntityTypeSearchRequest} request The search criteria and pagination information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchEntityTypes: function(request, callBack) {
      return this._start()
          .uri('/api/entity/type/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Searches the event logs with the specified criteria and pagination.
   *
   * @param {EventLogSearchRequest} request The search criteria and pagination information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchEventLogs: function(request, callBack) {
      return this._start()
          .uri('/api/system/event-log/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Searches the login records with the specified criteria and pagination.
   *
   * @param {LoginRecordSearchRequest} request The search criteria and pagination information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchLoginRecords: function(request, callBack) {
      return this._start()
          .uri('/api/system/login-record/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Retrieves the users for the given ids. If any id is invalid, it is ignored.
   *
   * @param {Array<string>} ids The user ids to search for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to searchUsersByIds, use that method instead.
   */
  searchUsers: function(ids, callBack) {
      return this._start()
          .uri('/api/user/search')
          .urlParameter('ids', ids)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the users for the given ids. If any id is invalid, it is ignored.
   *
   * @param {Array<string>} ids The user ids to search for.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchUsersByIds: function(ids, callBack) {
      return this._start()
          .uri('/api/user/search')
          .urlParameter('ids', ids)
          .get()
          .go(callBack);
  },

  /**
   * Retrieves the users for the given search criteria and pagination.
   *
   * @param {SearchRequest} request The search criteria and pagination constraints. Fields used: ids, query, queryString, numberOfResults, orderBy, startRow,
   *    and sortFields.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  searchUsersByQuery: function(request, callBack) {
      return this._start()
          .uri('/api/user/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Retrieves the users for the given search criteria and pagination.
   *
   * @param {SearchRequest} request The search criteria and pagination constraints. Fields used: ids, query, queryString, numberOfResults, orderBy, startRow,
   *    and sortFields.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to searchUsersByQuery, use that method instead.
   */
  searchUsersByQueryString: function(request, callBack) {
      return this._start()
          .uri('/api/user/search')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Send an email using an email template id. You can optionally provide <code>requestData</code> to access key value
   * pairs in the email template.
   *
   * @param {UUIDString} emailTemplateId The id for the template.
   * @param {SendRequest} request The send email request that contains all of the information used to send the email.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  sendEmail: function(emailTemplateId, request, callBack) {
      return this._start()
          .uri('/api/email/send')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Sends out an email to a parent that they need to register and create a family or need to log in and add a child to their existing family.
   *
   * @param {FamilyEmailRequest} request The request object that contains the parent email.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  sendFamilyRequestEmail: function(request, callBack) {
      return this._start()
          .uri('/api/user/family/request')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Send a passwordless authentication code in an email to complete login.
   *
   * @param {PasswordlessSendRequest} request The passwordless send request that contains all of the information used to send an email containing a code.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  sendPasswordlessCode: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/passwordless/send')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Send a Two Factor authentication code to assist in setting up Two Factor authentication or disabling.
   *
   * @param {TwoFactorSendRequest} request The request object that contains all of the information used to send the code.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to sendTwoFactorCodeForEnableDisable, use that method instead.
   */
  sendTwoFactorCode: function(request, callBack) {
      return this._start()
          .uri('/api/two-factor/send')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Send a Two Factor authentication code to assist in setting up Two Factor authentication or disabling.
   *
   * @param {TwoFactorSendRequest} request The request object that contains all of the information used to send the code.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  sendTwoFactorCodeForEnableDisable: function(request, callBack) {
      return this._start()
          .uri('/api/two-factor/send')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Send a Two Factor authentication code to allow the completion of Two Factor authentication.
   *
   * @param {string} twoFactorId The Id returned by the Login API necessary to complete Two Factor authentication.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to sendTwoFactorCodeForLoginUsingMethod, use that method instead.
   */
  sendTwoFactorCodeForLogin: function(twoFactorId, callBack) {
      return this._startAnonymous()
          .header('Content-Type', 'text/plain')
          .uri('/api/two-factor/send')
          .urlSegment(twoFactorId)
          .post()
          .go(callBack);
  },

  /**
   * Send a Two Factor authentication code to allow the completion of Two Factor authentication.
   *
   * @param {string} twoFactorId The Id returned by the Login API necessary to complete Two Factor authentication.
   * @param {TwoFactorSendRequest} request The Two Factor send request that contains all of the information used to send the Two Factor code to the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  sendTwoFactorCodeForLoginUsingMethod: function(twoFactorId, request, callBack) {
      return this._startAnonymous()
          .uri('/api/two-factor/send')
          .urlSegment(twoFactorId)
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Begins a login request for a 3rd party login that requires user interaction such as HYPR.
   *
   * @param {IdentityProviderStartLoginRequest} request The third-party login request that contains information from the third-party login
   *    providers that FusionAuth uses to reconcile the user's account.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  startIdentityProviderLogin: function(request, callBack) {
      return this._start()
          .uri('/api/identity-provider/start')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Start a passwordless login request by generating a passwordless code. This code can be sent to the User using the Send
   * Passwordless Code API or using a mechanism outside of FusionAuth. The passwordless login is completed by using the Passwordless Login API with this code.
   *
   * @param {PasswordlessStartRequest} request The passwordless start request that contains all of the information used to begin the passwordless login request.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  startPasswordlessLogin: function(request, callBack) {
      return this._start()
          .uri('/api/passwordless/start')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Start a Two-Factor login request by generating a two-factor identifier. This code can then be sent to the Two Factor Send 
   * API (/api/two-factor/send)in order to send a one-time use code to a user. You can also use one-time use code returned 
   * to send the code out-of-band. The Two-Factor login is completed by making a request to the Two-Factor Login 
   * API (/api/two-factor/login). with the two-factor identifier and the one-time use code.
   * 
   * This API is intended to allow you to begin a Two-Factor login outside of a normal login that originated from the Login API (/api/login).
   *
   * @param {TwoFactorStartRequest} request The Two-Factor start request that contains all of the information used to begin the Two-Factor login request.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  startTwoFactorLogin: function(request, callBack) {
      return this._start()
          .uri('/api/two-factor/start')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Complete login using a 2FA challenge
   *
   * @param {TwoFactorLoginRequest} request The login request that contains the user credentials used to log them in.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  twoFactorLogin: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/two-factor/login')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Updates an API key by given id
   *
   * @param {UUIDString} apiKeyId The Id of the API key to update.
   * @param {APIKeyRequest} request The request object that contains all of the information used to create the API Key.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateAPIKey: function(apiKeyId, request, callBack) {
      return this._start()
          .uri('/api/api-key')
          .urlSegment(apiKeyId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the application with the given Id.
   *
   * @param {UUIDString} applicationId The Id of the application to update.
   * @param {ApplicationRequest} request The request that contains all of the new application information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateApplication: function(applicationId, request, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the application role with the given id for the application.
   *
   * @param {UUIDString} applicationId The Id of the application that the role belongs to.
   * @param {UUIDString} roleId The Id of the role to update.
   * @param {ApplicationRequest} request The request that contains all of the new role information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateApplicationRole: function(applicationId, roleId, request, callBack) {
      return this._start()
          .uri('/api/application')
          .urlSegment(applicationId)
          .urlSegment("role")
          .urlSegment(roleId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the connector with the given Id.
   *
   * @param {UUIDString} connectorId The Id of the connector to update.
   * @param {ConnectorRequest} request The request object that contains all of the new connector information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateConnector: function(connectorId, request, callBack) {
      return this._start()
          .uri('/api/connector')
          .urlSegment(connectorId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the consent with the given Id.
   *
   * @param {UUIDString} consentId The Id of the consent to update.
   * @param {ConsentRequest} request The request that contains all of the new consent information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateConsent: function(consentId, request, callBack) {
      return this._start()
          .uri('/api/consent')
          .urlSegment(consentId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the email template with the given Id.
   *
   * @param {UUIDString} emailTemplateId The Id of the email template to update.
   * @param {EmailTemplateRequest} request The request that contains all of the new email template information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateEmailTemplate: function(emailTemplateId, request, callBack) {
      return this._start()
          .uri('/api/email/template')
          .urlSegment(emailTemplateId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the Entity with the given Id.
   *
   * @param {UUIDString} entityId The Id of the Entity to update.
   * @param {EntityRequest} request The request that contains all of the new Entity information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateEntity: function(entityId, request, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the Entity Type with the given Id.
   *
   * @param {UUIDString} entityTypeId The Id of the Entity Type to update.
   * @param {EntityTypeRequest} request The request that contains all of the new Entity Type information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateEntityType: function(entityTypeId, request, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the permission with the given id for the entity type.
   *
   * @param {UUIDString} entityTypeId The Id of the entityType that the permission belongs to.
   * @param {UUIDString} permissionId The Id of the permission to update.
   * @param {EntityTypeRequest} request The request that contains all of the new permission information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateEntityTypePermission: function(entityTypeId, permissionId, request, callBack) {
      return this._start()
          .uri('/api/entity/type')
          .urlSegment(entityTypeId)
          .urlSegment("permission")
          .urlSegment(permissionId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the form with the given Id.
   *
   * @param {UUIDString} formId The Id of the form to update.
   * @param {FormRequest} request The request object that contains all of the new form information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateForm: function(formId, request, callBack) {
      return this._start()
          .uri('/api/form')
          .urlSegment(formId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the form field with the given Id.
   *
   * @param {UUIDString} fieldId The Id of the form field to update.
   * @param {FormFieldRequest} request The request object that contains all of the new form field information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateFormField: function(fieldId, request, callBack) {
      return this._start()
          .uri('/api/form/field')
          .urlSegment(fieldId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the group with the given Id.
   *
   * @param {UUIDString} groupId The Id of the group to update.
   * @param {GroupRequest} request The request that contains all of the new group information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateGroup: function(groupId, request, callBack) {
      return this._start()
          .uri('/api/group')
          .urlSegment(groupId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the identity provider with the given Id.
   *
   * @param {UUIDString} identityProviderId The Id of the identity provider to update.
   * @param {IdentityProviderRequest} request The request object that contains the updated identity provider.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateIdentityProvider: function(identityProviderId, request, callBack) {
      return this._start()
          .uri('/api/identity-provider')
          .urlSegment(identityProviderId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the available integrations.
   *
   * @param {IntegrationRequest} request The request that contains all of the new integration information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateIntegrations: function(request, callBack) {
      return this._start()
          .uri('/api/integration')
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the key with the given Id.
   *
   * @param {UUIDString} keyId The Id of the key to update.
   * @param {KeyRequest} request The request that contains all of the new key information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateKey: function(keyId, request, callBack) {
      return this._start()
          .uri('/api/key')
          .urlSegment(keyId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the lambda with the given Id.
   *
   * @param {UUIDString} lambdaId The Id of the lambda to update.
   * @param {LambdaRequest} request The request that contains all of the new lambda information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateLambda: function(lambdaId, request, callBack) {
      return this._start()
          .uri('/api/lambda')
          .urlSegment(lambdaId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the message template with the given Id.
   *
   * @param {UUIDString} messageTemplateId The Id of the message template to update.
   * @param {MessageTemplateRequest} request The request that contains all of the new message template information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateMessageTemplate: function(messageTemplateId, request, callBack) {
      return this._start()
          .uri('/api/message/template')
          .urlSegment(messageTemplateId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the messenger with the given Id.
   *
   * @param {UUIDString} messengerId The Id of the messenger to update.
   * @param {MessengerRequest} request The request object that contains all of the new messenger information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateMessenger: function(messengerId, request, callBack) {
      return this._start()
          .uri('/api/messenger')
          .urlSegment(messengerId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the registration for the user with the given id and the application defined in the request.
   *
   * @param {UUIDString} userId The Id of the user whose registration is going to be updated.
   * @param {RegistrationRequest} request The request that contains all of the new registration information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateRegistration: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user/registration')
          .urlSegment(userId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the system configuration.
   *
   * @param {SystemConfigurationRequest} request The request that contains all of the new system configuration information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateSystemConfiguration: function(request, callBack) {
      return this._start()
          .uri('/api/system-configuration')
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the tenant with the given Id.
   *
   * @param {UUIDString} tenantId The Id of the tenant to update.
   * @param {TenantRequest} request The request that contains all of the new tenant information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateTenant: function(tenantId, request, callBack) {
      return this._start()
          .uri('/api/tenant')
          .urlSegment(tenantId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the theme with the given Id.
   *
   * @param {UUIDString} themeId The Id of the theme to update.
   * @param {ThemeRequest} request The request that contains all of the new theme information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateTheme: function(themeId, request, callBack) {
      return this._start()
          .uri('/api/theme')
          .urlSegment(themeId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the user with the given Id.
   *
   * @param {UUIDString} userId The Id of the user to update.
   * @param {UserRequest} request The request that contains all of the new user information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateUser: function(userId, request, callBack) {
      return this._start()
          .uri('/api/user')
          .urlSegment(userId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the user action with the given Id.
   *
   * @param {UUIDString} userActionId The Id of the user action to update.
   * @param {UserActionRequest} request The request that contains all of the new user action information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateUserAction: function(userActionId, request, callBack) {
      return this._start()
          .uri('/api/user-action')
          .urlSegment(userActionId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the user action reason with the given Id.
   *
   * @param {UUIDString} userActionReasonId The Id of the user action reason to update.
   * @param {UserActionReasonRequest} request The request that contains all of the new user action reason information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateUserActionReason: function(userActionReasonId, request, callBack) {
      return this._start()
          .uri('/api/user-action-reason')
          .urlSegment(userActionReasonId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates a single User consent by Id.
   *
   * @param {UUIDString} userConsentId The User Consent Id
   * @param {UserConsentRequest} request The request that contains the user consent information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateUserConsent: function(userConsentId, request, callBack) {
      return this._start()
          .uri('/api/user/consent')
          .urlSegment(userConsentId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Updates the webhook with the given Id.
   *
   * @param {UUIDString} webhookId The Id of the webhook to update.
   * @param {WebhookRequest} request The request that contains all of the new webhook information.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  updateWebhook: function(webhookId, request, callBack) {
      return this._start()
          .uri('/api/webhook')
          .urlSegment(webhookId)
          .setJSONBody(request)
          .put()
          .go(callBack);
  },

  /**
   * Creates or updates an Entity Grant. This is when a User/Entity is granted permissions to an Entity.
   *
   * @param {UUIDString} entityId The Id of the Entity that the User/Entity is being granted access to.
   * @param {EntityGrantRequest} request The request object that contains all of the information used to create the Entity Grant.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  upsertEntityGrant: function(entityId, request, callBack) {
      return this._start()
          .uri('/api/entity')
          .urlSegment(entityId)
          .urlSegment("grant")
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Validates the end-user provided user_code from the user-interaction of the Device Authorization Grant.
   * If you build your own activation form you should validate the user provided code prior to beginning the Authorization grant.
   *
   * @param {string} user_code The end-user verification code.
   * @param {string} client_id The client id.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  validateDevice: function(user_code, client_id, callBack) {
      return this._startAnonymous()
          .uri('/oauth2/device/validate')
          .urlParameter('user_code', user_code)
          .urlParameter('client_id', client_id)
          .get()
          .go(callBack);
  },

  /**
   * Validates the provided JWT (encoded JWT string) to ensure the token is valid. A valid access token is properly
   * signed and not expired.
   * <p>
   * This API may be used to verify the JWT as well as decode the encoded JWT into human readable identity claims.
   *
   * @param {string} encodedJWT The encoded JWT (access token).
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  validateJWT: function(encodedJWT, callBack) {
      return this._startAnonymous()
          .uri('/api/jwt/validate')
          .authorization('Bearer ' + encodedJWT)
          .get()
          .go(callBack);
  },

  /**
   * Confirms a email verification. The Id given is usually from an email sent to the user.
   *
   * @param {string} verificationId The email verification id sent to the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to verifyEmailAddress and changed to take a JSON request body, use that method instead.
   */
  verifyEmail: function(verificationId, callBack) {
      return this._startAnonymous()
          .header('Content-Type', 'text/plain')
          .uri('/api/user/verify-email')
          .urlSegment(verificationId)
          .post()
          .go(callBack);
  },

  /**
   * Confirms a user's email address. 
   * 
   * The request body will contain the verificationId. You may also be required to send a one-time use code based upon your configuration. When 
   * the tenant is configured to gate a user until their email address is verified, this procedures requires two values instead of one. 
   * The verificationId is a high entropy value and the one-time use code is a low entropy value that is easily entered in a user interactive form. The 
   * two values together are able to confirm a user's email address and mark the user's email address as verified.
   *
   * @param {VerifyEmailRequest} request The request that contains the verificationId and optional one-time use code paired with the verificationId.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  verifyEmailAddress: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/user/verify-email')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /**
   * Confirms an application registration. The Id given is usually from an email sent to the user.
   *
   * @param {string} verificationId The registration verification Id sent to the user.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   *
   * @deprecated This method has been renamed to verifyUserRegistration and changed to take a JSON request body, use that method instead.
   */
  verifyRegistration: function(verificationId, callBack) {
      return this._startAnonymous()
          .header('Content-Type', 'text/plain')
          .uri('/api/user/verify-registration')
          .urlSegment(verificationId)
          .post()
          .go(callBack);
  },

  /**
   * Confirms a user's registration. 
   * 
   * The request body will contain the verificationId. You may also be required to send a one-time use code based upon your configuration. When 
   * the application is configured to gate a user until their registration is verified, this procedures requires two values instead of one. 
   * The verificationId is a high entropy value and the one-time use code is a low entropy value that is easily entered in a user interactive form. The 
   * two values together are able to confirm a user's registration and mark the user's registration as verified.
   *
   * @param {VerifyRegistrationRequest} request The request that contains the verificationId and optional one-time use code paired with the verificationId.
   * @param {Function} callBack The response handler call back. This function will be passed the ClientResponse object.
   */
  verifyUserRegistration: function(request, callBack) {
      return this._startAnonymous()
          .uri('/api/user/verify-registration')
          .setJSONBody(request)
          .post()
          .go(callBack);
  },

  /* ===================================================================================================================
   * Private methods
   * ===================================================================================================================*/

  /**
   * creates a rest client
   *
   * @returns {RESTClient} The RESTClient that will be used to call.
   * @private
   */
  _start: function() {
    return this._startAnonymous().authorization(this.apiKey);
  },

  _startAnonymous: function() {
    let client = new RESTClient().setUrl(this.host);
    if (this.tenantId != null) {
      client.header('X-FusionAuth-TenantId', this.tenantId);
    }
    return client;
  }
};

/**
 * A 128 bit UUID in string format "8-4-4-4-12", for example "58D5E212-165B-4CA0-909B-C86B9CEE0111".
 *
 * @typedef {string} UUIDString
 */

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} AccessToken
 *
 * @property {string} [access_token]
 * @property {number} [expires_in]
 * @property {string} [id_token]
 * @property {string} [refresh_token]
 * @property {string} [scope]
 * @property {TokenType} [token_type]
 * @property {UUIDString} [userId]
 */


/**
 * @typedef {Object} ActionData
 *
 * @property {UUIDString} [actioneeUserId]
 * @property {UUIDString} [actionerUserId]
 * @property {Array<UUIDString>} [applicationIds]
 * @property {string} [comment]
 * @property {boolean} [emailUser]
 * @property {number} [expiry]
 * @property {boolean} [notifyUser]
 * @property {string} [option]
 * @property {UUIDString} [reasonId]
 * @property {UUIDString} [userActionId]
 */


/**
 * The user action request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ActionRequest
 *
 * @property {ActionData} [action]
 * @property {boolean} [broadcast]
 */


/**
 * The user action response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ActionResponse
 *
 * @property {UserActionLog} [action]
 * @property {Array<UserActionLog>} [actions]
 */


/**
 * Available JSON Web Algorithms (JWA) as described in RFC 7518 available for this JWT implementation.
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var Algorithm = {
  ES256: "ES256",
  ES384: "ES384",
  ES512: "ES512",
  HS256: "HS256",
  HS384: "HS384",
  HS512: "HS512",
  RS256: "RS256",
  RS384: "RS384",
  RS512: "RS512",
  none: "none"
};

/**
 * domain POJO to represent AuthenticationKey
 *
 * @author sanjay
 *
 * @typedef {Object} APIKey
 *
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {string} [key]
 * @property {boolean} [keyManager]
 * @property {number} [lastUpdateInstant]
 * @property {APIKeyMetaData} [metaData]
 * @property {APIKeyPermissions} [permissions]
 * @property {UUIDString} [tenantId]
 */


/**
 * @typedef {Object} APIKeyMetaData
 *
 * @property {Object<string, string>} [attributes]
 */


/**
 * @typedef {Object} APIKeyPermissions
 *
 * @property {Object<string, Object<string>>} [endpoints]
 */


/**
 * Authentication key request object.
 *
 * @author Sanjay
 *
 * @typedef {Object} APIKeyRequest
 *
 * @property {APIKey} [apiKey]
 * @property {UUIDString} [sourceKeyId]
 */


/**
 * Authentication key response object.
 *
 * @author Sanjay
 *
 * @typedef {Object} APIKeyResponse
 *
 * @property {APIKey} [apiKey]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} AppleApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {UUIDString} [keyId]
 * @property {string} [scope]
 * @property {string} [servicesId]
 * @property {string} [teamId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} AppleIdentityProvider
 * @extends BaseIdentityProvider<AppleApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {UUIDString} [keyId]
 * @property {string} [scope]
 * @property {string} [servicesId]
 * @property {string} [teamId]
 */


/**
 * @author Seth Musselman
 *
 * @typedef {Object} Application
 *
 * @property {boolean} [active]
 * @property {AuthenticationTokenConfiguration} [authenticationTokenConfiguration]
 * @property {CleanSpeakConfiguration} [cleanSpeakConfiguration]
 * @property {Object<string, Object>} [data]
 * @property {ApplicationEmailConfiguration} [emailConfiguration]
 * @property {ApplicationFormConfiguration} [formConfiguration]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {JWTConfiguration} [jwtConfiguration]
 * @property {LambdaConfiguration} [lambdaConfiguration]
 * @property {number} [lastUpdateInstant]
 * @property {LoginConfiguration} [loginConfiguration]
 * @property {ApplicationMultiFactorConfiguration} [multiFactorConfiguration]
 * @property {string} [name]
 * @property {OAuth2Configuration} [oauthConfiguration]
 * @property {PasswordlessConfiguration} [passwordlessConfiguration]
 * @property {RegistrationConfiguration} [registrationConfiguration]
 * @property {ApplicationRegistrationDeletePolicy} [registrationDeletePolicy]
 * @property {Array<ApplicationRole>} [roles]
 * @property {SAMLv2Configuration} [samlv2Configuration]
 * @property {ObjectState} [state]
 * @property {UUIDString} [tenantId]
 * @property {UUIDString} [themeId]
 * @property {RegistrationUnverifiedOptions} [unverified]
 * @property {UUIDString} [verificationEmailTemplateId]
 * @property {VerificationStrategy} [verificationStrategy]
 * @property {boolean} [verifyRegistration]
 */


/**
 * @typedef {Object} ApplicationEmailConfiguration
 *
 * @property {UUIDString} [emailVerificationEmailTemplateId]
 * @property {UUIDString} [forgotPasswordEmailTemplateId]
 * @property {UUIDString} [passwordlessEmailTemplateId]
 * @property {UUIDString} [setPasswordEmailTemplateId]
 */


/**
 * Events that are bound to applications.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ApplicationEvent
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ApplicationFormConfiguration
 *
 * @property {UUIDString} [adminRegistrationFormId]
 * @property {UUIDString} [selfServiceFormId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ApplicationMultiFactorConfiguration
 *
 * @property {MultiFactorEmailTemplate} [email]
 * @property {MultiFactorSMSTemplate} [sms]
 */


/**
 * A Application-level policy for deleting Users.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} ApplicationRegistrationDeletePolicy
 *
 * @property {TimeBasedDeletePolicy} [unverified]
 */


/**
 * The Application API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ApplicationRequest
 *
 * @property {Application} [application]
 * @property {ApplicationRole} [role]
 * @property {Array<UUIDString>} [webhookIds]
 */


/**
 * The Application API response.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ApplicationResponse
 *
 * @property {Application} [application]
 * @property {Array<Application>} [applications]
 * @property {ApplicationRole} [role]
 */


/**
 * A role given to a user for a specific application.
 *
 * @author Seth Musselman
 *
 * @typedef {Object} ApplicationRole
 *
 * @property {string} [description]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {boolean} [isDefault]
 * @property {boolean} [isSuperRole]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ApplicationUnverifiedConfiguration
 *
 * @property {UnverifiedBehavior} [registration]
 * @property {VerificationStrategy} [verificationStrategy]
 * @property {RegistrationUnverifiedOptions} [whenGated]
 */


/**
 * This class is a simple attachment with a byte array, name and MIME type.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Attachment
 *
 * @property {Array<number>} [attachment]
 * @property {string} [mime]
 * @property {string} [name]
 */


/**
 * An audit log.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} AuditLog
 *
 * @property {Object<string, Object>} [data]
 * @property {number} [id]
 * @property {number} [insertInstant]
 * @property {string} [insertUser]
 * @property {string} [message]
 * @property {Object} [newValue]
 * @property {Object} [oldValue]
 * @property {string} [reason]
 */


/**
 * @typedef {Object} AuditLogConfiguration
 *
 * @property {DeleteConfiguration} [delete]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} AuditLogExportRequest
 * @extends BaseExportRequest
 *
 * @property {AuditLogSearchCriteria} [criteria]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} AuditLogRequest
 *
 * @property {AuditLog} [auditLog]
 */


/**
 * Audit log response.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} AuditLogResponse
 *
 * @property {AuditLog} [auditLog]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} AuditLogSearchCriteria
 * @extends BaseSearchCriteria
 *
 * @property {number} [end]
 * @property {string} [message]
 * @property {number} [start]
 * @property {string} [user]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} AuditLogSearchRequest
 *
 * @property {AuditLogSearchCriteria} [search]
 */


/**
 * Audit log response.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} AuditLogSearchResponse
 *
 * @property {Array<AuditLog>} [auditLogs]
 * @property {number} [total]
 */


/**
 * @typedef {Object} AuthenticationTokenConfiguration
 * @extends Enableable
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} AuthenticatorConfiguration
 *
 * @property {TOTPAlgorithm} [algorithm]
 * @property {number} [codeLength]
 * @property {number} [timeStep]
 */


/**

 *
 * @typedef {Object} BaseConnectorConfiguration
 *
 * @property {Object<string, Object>} [data]
 * @property {boolean} [debug]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {ConnectorType} [type]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} BaseElasticSearchCriteria
 * @extends BaseSearchCriteria
 *
 * @property {boolean} [accurateTotal]
 * @property {Array<UUIDString>} [ids]
 * @property {string} [query]
 * @property {string} [queryString]
 * @property {Array<SortField>} [sortFields]
 */


/**
 * Base-class for all FusionAuth events.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} BaseEvent
 *
 * @property {number} [createInstant]
 * @property {UUIDString} [id]
 * @property {UUIDString} [tenantId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} BaseExportRequest
 *
 * @property {string} [dateTimeSecondsFormat]
 * @property {string} [zoneId]
 */


/**

 *
 * @typedef {Object} BaseIdentityProvider
 * @template {D}
 * @extends Enableable
 *
 * @property {Object<UUIDString, D>} [applicationConfiguration]
 * @property {Object<string, Object>} [data]
 * @property {boolean} [debug]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {LambdaConfiguration} [lambdaConfiguration]
 * @property {number} [lastUpdateInstant]
 * @property {IdentityProviderLinkingStrategy} [linkingStrategy]
 * @property {string} [name]
 * @property {IdentityProviderType} [type]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} BaseIdentityProviderApplicationConfiguration
 * @extends Enableable
 *
 * @property {boolean} [createRegistration]
 * @property {Object<string, Object>} [data]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} BaseLoginRequest
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [ipAddress]
 * @property {MetaData} [metaData]
 * @property {boolean} [noJWT]
 */


/**

 *
 * @typedef {Object} BaseMessengerConfiguration
 *
 * @property {Object<string, Object>} [data]
 * @property {boolean} [debug]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {string} [transport]
 * @property {MessengerType} [type]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} BaseSearchCriteria
 *
 * @property {number} [numberOfResults]
 * @property {string} [orderBy]
 * @property {number} [startRow]
 */


/**
 * @readonly
 * @enum
 */
var BreachAction = {
  Off: "Off",
  RecordOnly: "RecordOnly",
  NotifyUser: "NotifyUser",
  RequireChange: "RequireChange"
};

/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var BreachedPasswordStatus = {
  None: "None",
  ExactMatch: "ExactMatch",
  SubAddressMatch: "SubAddressMatch",
  PasswordOnly: "PasswordOnly",
  CommonPassword: "CommonPassword"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} BreachedPasswordTenantMetric
 *
 * @property {number} [actionRequired]
 * @property {number} [matchedCommonPasswordCount]
 * @property {number} [matchedExactCount]
 * @property {number} [matchedPasswordCount]
 * @property {number} [matchedSubAddressCount]
 * @property {number} [passwordsCheckedCount]
 */


/**
 * @readonly
 * @enum
 */
var BreachMatchMode = {
  Low: "Low",
  Medium: "Medium",
  High: "High"
};

/**
 * XML canonicalization method enumeration. This is used for the IdP and SP side of FusionAuth SAML.
 *
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var CanonicalizationMethod = {
  exclusive: "exclusive",
  exclusive_with_comments: "exclusive_with_comments",
  inclusive: "inclusive",
  inclusive_with_comments: "inclusive_with_comments"
};

/**
 * @typedef {Object} CertificateInformation
 *
 * @property {string} [issuer]
 * @property {string} [md5Fingerprint]
 * @property {string} [serialNumber]
 * @property {string} [sha1Fingerprint]
 * @property {string} [sha1Thumbprint]
 * @property {string} [sha256Fingerprint]
 * @property {string} [sha256Thumbprint]
 * @property {string} [subject]
 * @property {number} [validFrom]
 * @property {number} [validTo]
 */


/**
 * @author Trevor Smith
 *
 * @readonly
 * @enum
 */
var ChangePasswordReason = {
  Administrative: "Administrative",
  Breached: "Breached",
  Expired: "Expired",
  Validation: "Validation"
};

/**
 * Change password request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ChangePasswordRequest
 *
 * @property {string} [currentPassword]
 * @property {string} [loginId]
 * @property {string} [password]
 * @property {string} [refreshToken]
 */


/**
 * Change password response object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} ChangePasswordResponse
 *
 * @property {string} [oneTimePassword]
 * @property {Object<string, Object>} [state]
 */


/**
 * CleanSpeak configuration at the system and application level.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} CleanSpeakConfiguration
 * @extends Enableable
 *
 * @property {string} [apiKey]
 * @property {Array<UUIDString>} [applicationIds]
 * @property {string} [url]
 * @property {UsernameModeration} [usernameModeration]
 */


/**
 * @readonly
 * @enum
 */
var ClientAuthenticationMethod = {
  none: "none",
  client_secret_basic: "client_secret_basic",
  client_secret_post: "client_secret_post"
};

/**
 * @author Brett Guy
 *
 * @readonly
 * @enum
 */
var ClientAuthenticationPolicy = {
  Required: "Required",
  NotRequired: "NotRequired",
  NotRequiredWhenUsingPKCE: "NotRequiredWhenUsingPKCE"
};

/**
 * @author Trevor Smith
 *
 * @typedef {Object} ConnectorPolicy
 *
 * @property {UUIDString} [connectorId]
 * @property {Object<string, Object>} [data]
 * @property {Object<string>} [domains]
 * @property {boolean} [migrate]
 */


/**
 * @author Trevor Smith
 *
 * @typedef {Object} ConnectorRequest
 *
 * @property {BaseConnectorConfiguration} [connector]
 */


/**
 * @author Trevor Smith
 *
 * @typedef {Object} ConnectorResponse
 *
 * @property {BaseConnectorConfiguration} [connector]
 * @property {Array<BaseConnectorConfiguration>} [connectors]
 */


/**
 * The types of connectors. This enum is stored as an ordinal on the <code>identities</code> table, order must be maintained.
 *
 * @author Trevor Smith
 *
 * @readonly
 * @enum
 */
var ConnectorType = {
  FusionAuth: "FusionAuth",
  Generic: "Generic",
  LDAP: "LDAP"
};

/**
 * Models a consent.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} Consent
 *
 * @property {UUIDString} [consentEmailTemplateId]
 * @property {LocalizedIntegers} [countryMinimumAgeForSelfConsent]
 * @property {Object<string, Object>} [data]
 * @property {number} [defaultMinimumAgeForSelfConsent]
 * @property {EmailPlus} [emailPlus]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {boolean} [multipleValuesAllowed]
 * @property {string} [name]
 * @property {Array<string>} [values]
 */


/**
 * API request for User consent types.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} ConsentRequest
 *
 * @property {Consent} [consent]
 */


/**
 * API response for consent.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} ConsentResponse
 *
 * @property {Consent} [consent]
 * @property {Array<Consent>} [consents]
 */


/**
 * Models a consent.
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var ConsentStatus = {
  Active: "Active",
  Revoked: "Revoked"
};

/**
 * Status for content like usernames, profile attributes, etc.
 *
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var ContentStatus = {
  ACTIVE: "ACTIVE",
  PENDING: "PENDING",
  REJECTED: "REJECTED"
};

/**
 * @author Trevor Smith
 *
 * @typedef {Object} CORSConfiguration
 * @extends Enableable
 *
 * @property {boolean} [allowCredentials]
 * @property {Array<string>} [allowedHeaders]
 * @property {Array<HTTPMethod>} [allowedMethods]
 * @property {Array<string>} [allowedOrigins]
 * @property {boolean} [debug]
 * @property {Array<string>} [exposedHeaders]
 * @property {number} [preflightMaxAgeInSeconds]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} Count
 *
 * @property {number} [count]
 * @property {number} [interval]
 */


/**
 * Response for the daily active user report.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} DailyActiveUserReportResponse
 *
 * @property {Array<Count>} [dailyActiveUsers]
 * @property {number} [total]
 */


/**
 * Helper for dealing with default values.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} DefaultTools
 *
 */


/**
 * @typedef {Object} DeleteConfiguration
 * @extends Enableable
 *
 * @property {number} [numberOfDaysToRetain]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} DeviceInfo
 *
 * @property {string} [description]
 * @property {string} [lastAccessedAddress]
 * @property {number} [lastAccessedInstant]
 * @property {string} [name]
 * @property {DeviceType} [type]
 */


/**
 * @author Trevor Smith
 *
 * @typedef {Object} DeviceResponse
 *
 * @property {string} [device_code]
 * @property {number} [expires_in]
 * @property {number} [interval]
 * @property {string} [user_code]
 * @property {string} [verification_uri]
 * @property {string} [verification_uri_complete]
 */


/**
 * @readonly
 * @enum
 */
var DeviceType = {
  BROWSER: "BROWSER",
  DESKTOP: "DESKTOP",
  LAPTOP: "LAPTOP",
  MOBILE: "MOBILE",
  OTHER: "OTHER",
  SERVER: "SERVER",
  TABLET: "TABLET",
  TV: "TV",
  UNKNOWN: "UNKNOWN"
};

/**
 * A displayable raw login that includes application name and user loginId.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} DisplayableRawLogin
 * @extends RawLogin
 *
 * @property {string} [applicationName]
 * @property {string} [loginId]
 */


/**
 * Interface for all identity providers that can be domain based.
 *
 * @typedef {Object} DomainBasedIdentityProvider
 *
 */


/**
 * This class is an abstraction of a simple email message.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Email
 *
 * @property {Array<Attachment>} [attachments]
 * @property {Array<EmailAddress>} [bcc]
 * @property {Array<EmailAddress>} [cc]
 * @property {EmailAddress} [from]
 * @property {string} [html]
 * @property {EmailAddress} [replyTo]
 * @property {string} [subject]
 * @property {string} [text]
 * @property {Array<EmailAddress>} [to]
 */


/**
 * An email address.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EmailAddress
 *
 * @property {string} [address]
 * @property {string} [display]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} EmailConfiguration
 *
 * @property {string} [defaultFromEmail]
 * @property {string} [defaultFromName]
 * @property {UUIDString} [forgotPasswordEmailTemplateId]
 * @property {string} [host]
 * @property {string} [password]
 * @property {UUIDString} [passwordlessEmailTemplateId]
 * @property {number} [port]
 * @property {string} [properties]
 * @property {EmailSecurityType} [security]
 * @property {UUIDString} [setPasswordEmailTemplateId]
 * @property {EmailUnverifiedOptions} [unverified]
 * @property {string} [username]
 * @property {UUIDString} [verificationEmailTemplateId]
 * @property {VerificationStrategy} [verificationStrategy]
 * @property {boolean} [verifyEmail]
 * @property {boolean} [verifyEmailWhenChanged]
 */


/**
 * @typedef {Object} EmailPlus
 * @extends Enableable
 *
 * @property {UUIDString} [emailTemplateId]
 * @property {number} [maximumTimeToSendEmailInHours]
 * @property {number} [minimumTimeToSendEmailInHours]
 */


/**
 * @readonly
 * @enum
 */
var EmailSecurityType = {
  NONE: "NONE",
  SSL: "SSL",
  TLS: "TLS"
};

/**
 * Stores an email template used to send emails to users.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EmailTemplate
 *
 * @property {string} [defaultFromName]
 * @property {string} [defaultHtmlTemplate]
 * @property {string} [defaultSubject]
 * @property {string} [defaultTextTemplate]
 * @property {string} [fromEmail]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {LocalizedStrings} [localizedFromNames]
 * @property {LocalizedStrings} [localizedHtmlTemplates]
 * @property {LocalizedStrings} [localizedSubjects]
 * @property {LocalizedStrings} [localizedTextTemplates]
 * @property {string} [name]
 */


/**
 * @typedef {Object} EmailTemplateErrors
 *
 * @property {Object<string, string>} [parseErrors]
 * @property {Object<string, string>} [renderErrors]
 */


/**
 * Email template request.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EmailTemplateRequest
 *
 * @property {EmailTemplate} [emailTemplate]
 */


/**
 * Email template response.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EmailTemplateResponse
 *
 * @property {EmailTemplate} [emailTemplate]
 * @property {Array<EmailTemplate>} [emailTemplates]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} EmailUnverifiedOptions
 *
 * @property {boolean} [allowEmailChangeWhenGated]
 * @property {UnverifiedBehavior} [behavior]
 */


/**
 * Something that can be enabled and thus also disabled.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} Enableable
 *
 * @property {boolean} [enabled]
 */


/**
 * Models an entity that a user can be granted permissions to. Or an entity that can be granted permissions to another entity.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Entity
 *
 * @property {string} [clientId]
 * @property {string} [clientSecret]
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {UUIDString} [parentId]
 * @property {UUIDString} [tenantId]
 * @property {EntityType} [type]
 */


/**
 * A grant for an entity to a user or another entity.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityGrant
 *
 * @property {Object<string, Object>} [data]
 * @property {Entity} [entity]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {Object<string>} [permissions]
 * @property {UUIDString} [recipientEntityId]
 * @property {UUIDString} [userId]
 */


/**
 * Entity grant API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityGrantRequest
 *
 * @property {EntityGrant} [grant]
 */


/**
 * Entity grant API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityGrantResponse
 *
 * @property {EntityGrant} [grant]
 * @property {Array<EntityGrant>} [grants]
 */


/**
 * Search criteria for entity grants.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityGrantSearchCriteria
 * @extends BaseSearchCriteria
 *
 * @property {UUIDString} [entityId]
 * @property {string} [name]
 * @property {UUIDString} [userId]
 */


/**
 * Search request for entity grants.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityGrantSearchRequest
 *
 * @property {EntityGrantSearchCriteria} [search]
 */


/**
 * Search request for entity grants.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityGrantSearchResponse
 *
 * @property {Array<EntityGrant>} [grants]
 * @property {number} [total]
 */


/**
 * JWT Configuration for entities.
 *
 * @typedef {Object} EntityJWTConfiguration
 * @extends Enableable
 *
 * @property {UUIDString} [accessTokenKeyId]
 * @property {number} [timeToLiveInSeconds]
 */


/**
 * Entity API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityRequest
 *
 * @property {Entity} [entity]
 */


/**
 * Entity API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityResponse
 *
 * @property {Entity} [entity]
 */


/**
 * This class is the entity query. It provides a build pattern as well as public fields for use on forms and in actions.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntitySearchCriteria
 * @extends BaseElasticSearchCriteria
 *
 */


/**
 * Search request for entities
 *
 * @author Brett Guy
 *
 * @typedef {Object} EntitySearchRequest
 *
 * @property {EntitySearchCriteria} [search]
 */


/**
 * Search request for entities
 *
 * @author Brett Guy
 *
 * @typedef {Object} EntitySearchResponse
 *
 * @property {Array<Entity>} [entities]
 * @property {number} [total]
 */


/**
 * Models an entity type that has a specific set of permissions. These are global objects and can be used across tenants.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityType
 *
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {EntityJWTConfiguration} [jwtConfiguration]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {Array<EntityTypePermission>} [permissions]
 */


/**
 * Models a specific entity type permission. This permission can be granted to users or other entities.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityTypePermission
 *
 * @property {Object<string, Object>} [data]
 * @property {string} [description]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {boolean} [isDefault]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 */


/**
 * Entity Type API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityTypeRequest
 *
 * @property {EntityType} [entityType]
 * @property {EntityTypePermission} [permission]
 */


/**
 * Entity Type API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityTypeResponse
 *
 * @property {EntityType} [entityType]
 * @property {Array<EntityType>} [entityTypes]
 * @property {EntityTypePermission} [permission]
 */


/**
 * Search criteria for entity types.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityTypeSearchCriteria
 * @extends BaseSearchCriteria
 *
 * @property {string} [name]
 */


/**
 * Search request for entity types.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityTypeSearchRequest
 *
 * @property {EntityTypeSearchCriteria} [search]
 */


/**
 * Search request for entity types.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EntityTypeSearchResponse
 *
 * @property {Array<EntityType>} [entityTypes]
 * @property {number} [total]
 */


/**
 * @author Brett Pontarelli
 *
 * @typedef {Object} EpicGamesApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Epic gaming login provider.
 *
 * @author Brett Pontarelli
 *
 * @typedef {Object} EpicGamesIdentityProvider
 * @extends BaseIdentityProvider<EpicGamesApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Defines an error.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Error
 *
 * @property {string} [code]
 * @property {string} [message]
 */


/**
 * Standard error domain object that can also be used as the response from an API call.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Errors
 *
 * @property {Object<string, Array<Error>>} [fieldErrors]
 * @property {Array<Error>} [generalErrors]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} EventConfiguration
 *
 * @property {Object<EventType, EventConfigurationData>} [events]
 */


/**
 * @typedef {Object} EventConfigurationData
 * @extends Enableable
 *
 * @property {TransactionType} [transactionType]
 */


/**
 * Event log used internally by FusionAuth to help developers debug hooks, Webhooks, email templates, etc.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EventLog
 *
 * @property {number} [id]
 * @property {number} [insertInstant]
 * @property {string} [message]
 * @property {EventLogType} [type]
 */


/**
 * @typedef {Object} EventLogConfiguration
 *
 * @property {number} [numberToRetain]
 */


/**
 * Event log response.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} EventLogResponse
 *
 * @property {EventLog} [eventLog]
 */


/**
 * Search criteria for the event log.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EventLogSearchCriteria
 * @extends BaseSearchCriteria
 *
 * @property {number} [end]
 * @property {string} [message]
 * @property {number} [start]
 * @property {EventLogType} [type]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} EventLogSearchRequest
 *
 * @property {EventLogSearchCriteria} [search]
 */


/**
 * Event log response.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EventLogSearchResponse
 *
 * @property {Array<EventLog>} [eventLogs]
 * @property {number} [total]
 */


/**
 * Event Log Type
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var EventLogType = {
  Information: "Information",
  Debug: "Debug",
  Error: "Error"
};

/**
 * Container for the event information. This is the JSON that is sent from FusionAuth to webhooks.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} EventRequest
 *
 * @property {BaseEvent} [event]
 */


/**
 * Models the event types that FusionAuth produces.
 *
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var EventType = {
  UserDelete: "user.delete",
  UserCreate: "user.create",
  UserUpdate: "user.update",
  UserDeactivate: "user.deactivate",
  UserBulkCreate: "user.bulk.create",
  UserReactivate: "user.reactivate",
  UserAction: "user.action",
  JWTRefreshTokenRevoke: "jwt.refresh-token.revoke",
  JWTRefresh: "jwt.refresh",
  JWTPublicKeyUpdate: "jwt.public-key.update",
  UserLoginSuccess: "user.login.success",
  UserLoginFailed: "user.login.failed",
  UserRegistrationCreate: "user.registration.create",
  UserRegistrationUpdate: "user.registration.update",
  UserRegistrationDelete: "user.registration.delete",
  UserRegistrationVerified: "user.registration.verified",
  UserEmailVerified: "user.email.verified",
  UserPasswordBreach: "user.password.breach",
  Test: "test"
};

/**
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var ExpiryUnit = {
  MINUTES: "MINUTES",
  HOURS: "HOURS",
  DAYS: "DAYS",
  WEEKS: "WEEKS",
  MONTHS: "MONTHS",
  YEARS: "YEARS"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ExternalIdentifierConfiguration
 *
 * @property {number} [authorizationGrantIdTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [changePasswordIdGenerator]
 * @property {number} [changePasswordIdTimeToLiveInSeconds]
 * @property {number} [deviceCodeTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [deviceUserCodeIdGenerator]
 * @property {SecureGeneratorConfiguration} [emailVerificationIdGenerator]
 * @property {number} [emailVerificationIdTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [emailVerificationOneTimeCodeGenerator]
 * @property {number} [externalAuthenticationIdTimeToLiveInSeconds]
 * @property {number} [oneTimePasswordTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [passwordlessLoginGenerator]
 * @property {number} [passwordlessLoginTimeToLiveInSeconds]
 * @property {number} [pendingAccountLinkTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [registrationVerificationIdGenerator]
 * @property {number} [registrationVerificationIdTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [registrationVerificationOneTimeCodeGenerator]
 * @property {number} [samlv2AuthNRequestIdTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [setupPasswordIdGenerator]
 * @property {number} [setupPasswordIdTimeToLiveInSeconds]
 * @property {number} [twoFactorIdTimeToLiveInSeconds]
 * @property {SecureGeneratorConfiguration} [twoFactorOneTimeCodeIdGenerator]
 * @property {number} [twoFactorOneTimeCodeIdTimeToLiveInSeconds]
 * @property {number} [twoFactorTrustIdTimeToLiveInSeconds]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ExternalJWTApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 */


/**
 * External JWT-only identity provider.
 *
 * @author Daniel DeGroff and Brian Pontarelli
 *
 * @typedef {Object} ExternalJWTIdentityProvider
 * @extends BaseIdentityProvider<ExternalJWTApplicationConfiguration>
 *
 * @property {Object<string, string>} [claimMap]
 * @property {UUIDString} [defaultKeyId]
 * @property {Object<string>} [domains]
 * @property {string} [headerKeyParameter]
 * @property {IdentityProviderOauth2Configuration} [oauth2]
 * @property {string} [uniqueIdentityClaim]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} FacebookApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [appId]
 * @property {string} [buttonText]
 * @property {string} [client_secret]
 * @property {string} [fields]
 * @property {IdentityProviderLoginMethod} [loginMethod]
 * @property {string} [permissions]
 */


/**
 * Facebook social login provider.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} FacebookIdentityProvider
 * @extends BaseIdentityProvider<FacebookApplicationConfiguration>
 *
 * @property {string} [appId]
 * @property {string} [buttonText]
 * @property {string} [client_secret]
 * @property {string} [fields]
 * @property {IdentityProviderLoginMethod} [loginMethod]
 * @property {string} [permissions]
 */


/**
 * Configuration for the behavior of failed login attempts. This helps us protect against brute force password attacks.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} FailedAuthenticationConfiguration
 *
 * @property {number} [actionDuration]
 * @property {ExpiryUnit} [actionDurationUnit]
 * @property {number} [resetCountInSeconds]
 * @property {number} [tooManyAttempts]
 * @property {UUIDString} [userActionId]
 */


/**
 * Models a family grouping of users.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Family
 *
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {Array<FamilyMember>} [members]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} FamilyConfiguration
 * @extends Enableable
 *
 * @property {boolean} [allowChildRegistrations]
 * @property {UUIDString} [confirmChildEmailTemplateId]
 * @property {boolean} [deleteOrphanedAccounts]
 * @property {number} [deleteOrphanedAccountsDays]
 * @property {UUIDString} [familyRequestEmailTemplateId]
 * @property {number} [maximumChildAge]
 * @property {number} [minimumOwnerAge]
 * @property {boolean} [parentEmailRequired]
 * @property {UUIDString} [parentRegistrationEmailTemplateId]
 */


/**
 * API request for sending out family requests to parent's.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} FamilyEmailRequest
 *
 * @property {string} [parentEmail]
 */


/**
 * Models a single family member.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} FamilyMember
 *
 * @property {Object<string, Object>} [data]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {boolean} [owner]
 * @property {FamilyRole} [role]
 * @property {UUIDString} [userId]
 */


/**
 * API request for managing families and members.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} FamilyRequest
 *
 * @property {FamilyMember} [familyMember]
 */


/**
 * API response for managing families and members.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} FamilyResponse
 *
 * @property {Array<Family>} [families]
 * @property {Family} [family]
 */


/**
 * @readonly
 * @enum
 */
var FamilyRole = {
  Child: "Child",
  Teen: "Teen",
  Adult: "Adult"
};

/**
 * Forgot password request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ForgotPasswordRequest
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [changePasswordId]
 * @property {string} [email]
 * @property {string} [loginId]
 * @property {boolean} [sendForgotPasswordEmail]
 * @property {Object<string, Object>} [state]
 * @property {string} [username]
 */


/**
 * Forgot password response object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} ForgotPasswordResponse
 *
 * @property {string} [changePasswordId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} Form
 *
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {Array<FormStep>} [steps]
 * @property {FormType} [type]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var FormControl = {
  checkbox: "checkbox",
  number: "number",
  password: "password",
  radio: "radio",
  select: "select",
  textarea: "textarea",
  text: "text"
};

/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var FormDataType = {
  bool: "bool",
  consent: "consent",
  date: "date",
  email: "email",
  number: "number",
  string: "string"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} FormField
 *
 * @property {boolean} [confirm]
 * @property {UUIDString} [consentId]
 * @property {FormControl} [control]
 * @property {Object<string, Object>} [data]
 * @property {string} [description]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {string} [key]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {Array<string>} [options]
 * @property {boolean} [required]
 * @property {FormDataType} [type]
 * @property {FormFieldValidator} [validator]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var FormFieldAdminPolicy = {
  Edit: "Edit",
  View: "View"
};

/**
 * The FormField API request object.
 *
 * @author Brett Guy
 *
 * @typedef {Object} FormFieldRequest
 *
 * @property {FormField} [field]
 * @property {Array<FormField>} [fields]
 */


/**
 * Form field response.
 *
 * @author Brett Guy
 *
 * @typedef {Object} FormFieldResponse
 *
 * @property {FormField} [field]
 * @property {Array<FormField>} [fields]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} FormFieldValidator
 * @extends Enableable
 *
 * @property {string} [expression]
 */


/**
 * Form response.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} FormRequest
 *
 * @property {Form} [form]
 */


/**
 * Form response.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} FormResponse
 *
 * @property {Form} [form]
 * @property {Array<Form>} [forms]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} FormStep
 *
 * @property {Array<UUIDString>} [fields]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var FormType = {
  registration: "registration",
  adminRegistration: "adminRegistration",
  adminUser: "adminUser",
  selfServiceUser: "selfServiceUser"
};

/**
 * Models the FusionAuth connector.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} FusionAuthConnectorConfiguration
 * @extends BaseConnectorConfiguration
 *
 */


/**
 * Models a generic connector.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} GenericConnectorConfiguration
 * @extends BaseConnectorConfiguration
 *
 * @property {string} [authenticationURL]
 * @property {number} [connectTimeout]
 * @property {HTTPHeaders} [headers]
 * @property {string} [httpAuthenticationPassword]
 * @property {string} [httpAuthenticationUsername]
 * @property {number} [readTimeout]
 * @property {UUIDString} [sslCertificateKeyId]
 */


/**
 * @author Brett Guy
 *
 * @typedef {Object} GenericMessengerConfiguration
 * @extends BaseMessengerConfiguration
 *
 * @property {number} [connectTimeout]
 * @property {HTTPHeaders} [headers]
 * @property {string} [httpAuthenticationPassword]
 * @property {string} [httpAuthenticationUsername]
 * @property {number} [readTimeout]
 * @property {string} [sslCertificate]
 * @property {string} [url]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} GoogleApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {IdentityProviderLoginMethod} [loginMethod]
 * @property {string} [scope]
 */


/**
 * Google social login provider.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} GoogleIdentityProvider
 * @extends BaseIdentityProvider<GoogleApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {IdentityProviderLoginMethod} [loginMethod]
 * @property {string} [scope]
 */


/**
 * Authorization Grant types as defined by the <a href="https://tools.ietf.org/html/rfc6749">The OAuth 2.0 Authorization
 * Framework - RFC 6749</a>.
 * <p>
 * Specific names as defined by <a href="https://tools.ietf.org/html/rfc7591#section-4.1">
 * OAuth 2.0 Dynamic Client Registration Protocol - RFC 7591 Section 4.1</a>
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var GrantType = {
  authorization_code: "authorization_code",
  implicit: "implicit",
  password: "password",
  client_credentials: "client_credentials",
  refresh_token: "refresh_token",
  unknown: "unknown",
  device_code: "urn:ietf:params:oauth:grant-type:device_code"
};

/**
 * @author Tyler Scott
 *
 * @typedef {Object} Group
 *
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {Object<UUIDString, Array<ApplicationRole>>} [roles]
 * @property {UUIDString} [tenantId]
 */


/**
 * A User's membership into a Group
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} GroupMember
 *
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [groupId]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {UUIDString} [userId]
 */


/**
 * Group API request object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} GroupRequest
 *
 * @property {Group} [group]
 * @property {Array<UUIDString>} [roleIds]
 */


/**
 * Group API response object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} GroupResponse
 *
 * @property {Group} [group]
 * @property {Array<Group>} [groups]
 */


/**
 * @typedef {Object} HistoryItem
 *
 * @property {UUIDString} [actionerUserId]
 * @property {string} [comment]
 * @property {number} [createInstant]
 * @property {number} [expiry]
 */


/**
 * Type for webhook headers.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} HTTPHeaders
 * @extends Object<string, string>
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var HTTPMethod = {
  GET: "GET",
  POST: "POST",
  PUT: "PUT",
  DELETE: "DELETE",
  HEAD: "HEAD",
  OPTIONS: "OPTIONS",
  PATCH: "PATCH"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} HYPRApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [relyingPartyApplicationId]
 * @property {string} [relyingPartyURL]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} HYPRIdentityProvider
 * @extends BaseIdentityProvider<HYPRApplicationConfiguration>
 *
 * @property {string} [relyingPartyApplicationId]
 * @property {string} [relyingPartyURL]
 */


/**
 * @typedef {Object} IdentityProviderDetails
 *
 * @property {Array<UUIDString>} [applicationIds]
 * @property {UUIDString} [id]
 * @property {string} [idpEndpoint]
 * @property {string} [name]
 * @property {IdentityProviderOauth2Configuration} [oauth2]
 * @property {IdentityProviderType} [type]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderLink
 *
 * @property {Object<string, Object>} [data]
 * @property {string} [displayName]
 * @property {UUIDString} [identityProviderId]
 * @property {string} [identityProviderUserId]
 * @property {number} [insertInstant]
 * @property {number} [lastLoginInstant]
 * @property {UUIDString} [tenantId]
 * @property {string} [token]
 * @property {UUIDString} [userId]
 */


/**
 * The IdP behavior when no user link has been made yet.
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var IdentityProviderLinkingStrategy = {
  CreatePendingLink: "CreatePendingLink",
  LinkAnonymously: "LinkAnonymously",
  LinkByEmail: "LinkByEmail",
  LinkByEmailForExistingUser: "LinkByEmailForExistingUser",
  LinkByUsername: "LinkByUsername",
  LinkByUsernameForExistingUser: "LinkByUsernameForExistingUser",
  Unsupported: "Unsupported"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderLinkRequest
 *
 * @property {string} [displayName]
 * @property {UUIDString} [identityProviderId]
 * @property {string} [identityProviderUserId]
 * @property {string} [pendingIdPLinkId]
 * @property {UUIDString} [userId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderLinkResponse
 *
 * @property {IdentityProviderLink} [identityProviderLink]
 * @property {Array<IdentityProviderLink>} [identityProviderLinks]
 */


/**
 * @author Brett Pontarelli
 *
 * @readonly
 * @enum
 */
var IdentityProviderLoginMethod = {
  UsePopup: "UsePopup",
  UseRedirect: "UseRedirect"
};

/**
 * Login API request object used for login to third-party systems (i.e. Login with Facebook).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} IdentityProviderLoginRequest
 * @extends BaseLoginRequest
 *
 * @property {Object<string, string>} [data]
 * @property {string} [encodedJWT]
 * @property {UUIDString} [identityProviderId]
 * @property {boolean} [noLink]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderOauth2Configuration
 *
 * @property {string} [authorization_endpoint]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {ClientAuthenticationMethod} [clientAuthenticationMethod]
 * @property {string} [emailClaim]
 * @property {string} [issuer]
 * @property {string} [scope]
 * @property {string} [token_endpoint]
 * @property {string} [uniqueIdClaim]
 * @property {string} [userinfo_endpoint]
 * @property {string} [usernameClaim]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderRequest
 *
 * @property {BaseIdentityProvider<Object>} [identityProvider]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderResponse
 *
 * @property {BaseIdentityProvider<Object>} [identityProvider]
 * @property {Array<BaseIdentityProvider<Object>>} [identityProviders]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderStartLoginRequest
 * @extends BaseLoginRequest
 *
 * @property {Object<string, string>} [data]
 * @property {UUIDString} [identityProviderId]
 * @property {string} [loginId]
 * @property {Object<string, Object>} [state]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IdentityProviderStartLoginResponse
 *
 * @property {string} [code]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var IdentityProviderType = {
  Apple: "Apple",
  EpicGames: "EpicGames",
  ExternalJWT: "ExternalJWT",
  Facebook: "Facebook",
  Google: "Google",
  HYPR: "HYPR",
  LinkedIn: "LinkedIn",
  Nintendo: "Nintendo",
  OpenIDConnect: "OpenIDConnect",
  SAMLv2: "SAMLv2",
  SAMLv2IdPInitiated: "SAMLv2IdPInitiated",
  SonyPSN: "SonyPSN",
  Steam: "Steam",
  Twitch: "Twitch",
  Twitter: "Twitter",
  Xbox: "Xbox"
};

/**
 * Import request.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ImportRequest
 *
 * @property {string} [encryptionScheme]
 * @property {number} [factor]
 * @property {Array<User>} [users]
 * @property {boolean} [validateDbConstraints]
 */


/**
 * The Integration Request
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} IntegrationRequest
 *
 * @property {Integrations} [integrations]
 */


/**
 * The Integration Response
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} IntegrationResponse
 *
 * @property {Integrations} [integrations]
 */


/**
 * Available Integrations
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} Integrations
 *
 * @property {CleanSpeakConfiguration} [cleanspeak]
 * @property {KafkaConfiguration} [kafka]
 */


/**
 * Counts for a period.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} IntervalCount
 *
 * @property {UUIDString} [applicationId]
 * @property {number} [count]
 * @property {number} [decrementedCount]
 * @property {number} [period]
 */


/**
 * A user over an period (for daily and monthly active user calculations).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} IntervalUser
 *
 * @property {UUIDString} [applicationId]
 * @property {number} [period]
 * @property {UUIDString} [userId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IntrospectResponse
 * @extends Object<string, Object>
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} IssueResponse
 *
 * @property {string} [refreshToken]
 * @property {string} [token]
 */


/**
 * A JSON Web Key as defined by <a href="https://tools.ietf.org/html/rfc7517#section-4">RFC 7517 JSON Web Key (JWK)
 * Section 4</a> and <a href="https://tools.ietf.org/html/rfc7518">RFC 7518 JSON Web Algorithms (JWA)</a>.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} JSONWebKey
 *
 * @property {Algorithm} [alg]
 * @property {string} [crv]
 * @property {string} [d]
 * @property {string} [dp]
 * @property {string} [dq]
 * @property {string} [e]
 * @property {string} [kid]
 * @property {KeyType} [kty]
 * @property {string} [n]
 * @property {string} [p]
 * @property {string} [q]
 * @property {string} [qi]
 * @property {string} [use]
 * @property {string} [x]
 * @property {Array<string>} [x5c]
 * @property {string} [x5t]
 * @property {string} [x5t#S256]
 * @property {string} [y]
 */


/**
 * Interface for any object that can provide JSON Web key Information.
 *
 * @typedef {Object} JSONWebKeyInfoProvider
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} JWKSResponse
 *
 * @property {Array<JSONWebKey>} [keys]
 */


/**
 * JSON Web Token (JWT) as defined by RFC 7519.
 * <pre>
 * From RFC 7519 Section 1. Introduction:
 *    The suggested pronunciation of JWT is the same as the English word "jot".
 * </pre>
 * The JWT is not Thread-Safe and should not be re-used.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} JWT | object
 *
 * @property {Object} [aud]
 * @property {number} [exp]
 * @property {number} [iat]
 * @property {string} [iss]
 * @property {string} [jti]
 * @property {number} [nbf]
 * @property {string} [sub]
 */


/**
 * JWT Configuration. A JWT Configuration for an Application may not be active if it is using the global configuration, the configuration
 * may be <code>enabled = false</code>.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} JWTConfiguration
 * @extends Enableable
 *
 * @property {UUIDString} [accessTokenKeyId]
 * @property {UUIDString} [idTokenKeyId]
 * @property {RefreshTokenExpirationPolicy} [refreshTokenExpirationPolicy]
 * @property {RefreshTokenRevocationPolicy} [refreshTokenRevocationPolicy]
 * @property {number} [refreshTokenTimeToLiveInMinutes]
 * @property {RefreshTokenUsagePolicy} [refreshTokenUsagePolicy]
 * @property {number} [timeToLiveInSeconds]
 */


/**
 * Models the JWT public key Refresh Token Revoke Event (and can be converted to JSON). This event might be for a single
 * token, a user or an entire application.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} JWTPublicKeyUpdateEvent
 * @extends BaseEvent
 *
 * @property {Object<UUIDString>} [applicationIds]
 */


/**
 * Models the JWT Refresh Event. This event will be fired when a JWT is "refreshed" (generated) using a Refresh Token.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} JWTRefreshEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [original]
 * @property {string} [refreshToken]
 * @property {string} [token]
 * @property {UUIDString} [userId]
 */


/**
 * API response for refreshing a JWT with a Refresh Token.
 * <p>
 * Using a different response object from RefreshTokenResponse because the retrieve response will return an object for refreshToken, and this is a string.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} JWTRefreshResponse
 *
 * @property {string} [refreshToken]
 * @property {string} [token]
 */


/**
 * Models the Refresh Token Revoke Event (and can be converted to JSON). This event might be for a single token, a user
 * or an entire application.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} JWTRefreshTokenRevokeEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {Object<UUIDString, number>} [applicationTimeToLiveInSeconds]
 * @property {User} [user]
 * @property {UUIDString} [userId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} KafkaConfiguration
 * @extends Enableable
 *
 * @property {string} [defaultTopic]
 * @property {Object<string, string>} [producer]
 */


/**
 * @author Brett Guy
 *
 * @typedef {Object} KafkaMessengerConfiguration
 * @extends BaseMessengerConfiguration
 *
 * @property {string} [defaultTopic]
 * @property {Object<string, string>} [producer]
 */


/**
 * Domain for a public key, key pair or an HMAC secret. This is used by KeyMaster to manage keys for JWTs, SAML, etc.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Key
 *
 * @property {KeyAlgorithm} [algorithm]
 * @property {string} [certificate]
 * @property {CertificateInformation} [certificateInformation]
 * @property {number} [expirationInstant]
 * @property {boolean} [hasPrivateKey]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {string} [issuer]
 * @property {string} [kid]
 * @property {number} [lastUpdateInstant]
 * @property {number} [length]
 * @property {string} [name]
 * @property {string} [privateKey]
 * @property {string} [publicKey]
 * @property {string} [secret]
 * @property {KeyType} [type]
 */


/**
 * @readonly
 * @enum
 */
var KeyAlgorithm = {
  ES256: "ES256",
  ES384: "ES384",
  ES512: "ES512",
  HS256: "HS256",
  HS384: "HS384",
  HS512: "HS512",
  RS256: "RS256",
  RS384: "RS384",
  RS512: "RS512"
};

/**
 * Key API request object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} KeyRequest
 *
 * @property {Key} [key]
 */


/**
 * Key API response object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} KeyResponse
 *
 * @property {Key} [key]
 * @property {Array<Key>} [keys]
 */


/**
 * @readonly
 * @enum
 */
var KeyType = {
  EC: "EC",
  RSA: "RSA",
  HMAC: "HMAC"
};

/**
 * The use type of a key.
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var KeyUse = {
  SignOnly: "SignOnly",
  SignAndVerify: "SignAndVerify",
  VerifyOnly: "VerifyOnly"
};

/**
 * A JavaScript lambda function that is executed during certain events inside FusionAuth.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Lambda
 * @extends Enableable
 *
 * @property {string} [body]
 * @property {boolean} [debug]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {LambdaType} [type]
 */


/**
 * @typedef {Object} LambdaConfiguration
 *
 * @property {UUIDString} [accessTokenPopulateId]
 * @property {UUIDString} [idTokenPopulateId]
 * @property {UUIDString} [samlv2PopulateId]
 */


/**
 * @typedef {Object} LambdaConfiguration
 *
 * @property {UUIDString} [reconcileId]
 */


/**
 * @typedef {Object} LambdaConfiguration
 *
 * @property {UUIDString} [reconcileId]
 */


/**
 * Lambda API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} LambdaRequest
 *
 * @property {Lambda} [lambda]
 */


/**
 * Lambda API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} LambdaResponse
 *
 * @property {Lambda} [lambda]
 * @property {Array<Lambda>} [lambdas]
 */


/**
 * The types of lambdas that indicate how they are invoked by FusionAuth.
 *
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var LambdaType = {
  JWTPopulate: "JWTPopulate",
  OpenIDReconcile: "OpenIDReconcile",
  SAMLv2Reconcile: "SAMLv2Reconcile",
  SAMLv2Populate: "SAMLv2Populate",
  AppleReconcile: "AppleReconcile",
  ExternalJWTReconcile: "ExternalJWTReconcile",
  FacebookReconcile: "FacebookReconcile",
  GoogleReconcile: "GoogleReconcile",
  HYPRReconcile: "HYPRReconcile",
  TwitterReconcile: "TwitterReconcile",
  LDAPConnectorReconcile: "LDAPConnectorReconcile",
  LinkedInReconcile: "LinkedInReconcile",
  EpicGamesReconcile: "EpicGamesReconcile",
  NintendoReconcile: "NintendoReconcile",
  SonyPSNReconcile: "SonyPSNReconcile",
  SteamReconcile: "SteamReconcile",
  TwitchReconcile: "TwitchReconcile",
  XboxReconcile: "XboxReconcile",
  ClientCredentialsJWTPopulate: "ClientCredentialsJWTPopulate"
};

/**
 * Models an LDAP connector.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} LDAPConnectorConfiguration
 * @extends BaseConnectorConfiguration
 *
 * @property {string} [authenticationURL]
 * @property {string} [baseStructure]
 * @property {number} [connectTimeout]
 * @property {string} [identifyingAttribute]
 * @property {LambdaConfiguration} [lambdaConfiguration]
 * @property {string} [loginIdAttribute]
 * @property {number} [readTimeout]
 * @property {Array<string>} [requestedAttributes]
 * @property {LDAPSecurityMethod} [securityMethod]
 * @property {string} [systemAccountDN]
 * @property {string} [systemAccountPassword]
 */


/**
 * @readonly
 * @enum
 */
var LDAPSecurityMethod = {
  None: "None",
  LDAPS: "LDAPS",
  StartTLS: "StartTLS"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} LinkedInApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} LinkedInIdentityProvider
 * @extends BaseIdentityProvider<LinkedInApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Models a set of localized Integers that can be stored as JSON.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} LocalizedIntegers
 * @extends Object<string, number>
 *
 */


/**
 * Models a set of localized Strings that can be stored as JSON.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} LocalizedStrings
 * @extends Object<string, string>
 *
 */


/**
 * A historical state of a user log event. Since events can be modified, this stores the historical state.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} LogHistory
 *
 * @property {Array<HistoryItem>} [historyItems]
 */


/**
 * @typedef {Object} LoginConfiguration
 *
 * @property {boolean} [allowTokenRefresh]
 * @property {boolean} [generateRefreshTokens]
 * @property {boolean} [requireAuthentication]
 */


/**
 * @readonly
 * @enum
 */
var LoginIdType = {
  email: "email",
  username: "username"
};

/**
 * The summary of the action that is preventing login to be returned on the login response.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} LoginPreventedResponse
 *
 * @property {UUIDString} [actionerUserId]
 * @property {UUIDString} [actionId]
 * @property {number} [expiry]
 * @property {string} [localizedName]
 * @property {string} [localizedOption]
 * @property {string} [localizedReason]
 * @property {string} [name]
 * @property {string} [option]
 * @property {string} [reason]
 * @property {string} [reasonCode]
 */


/**
 * @typedef {Object} LoginRecordConfiguration
 *
 * @property {DeleteConfiguration} [delete]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} LoginRecordExportRequest
 * @extends BaseExportRequest
 *
 * @property {LoginRecordSearchCriteria} [criteria]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} LoginRecordSearchCriteria
 * @extends BaseSearchCriteria
 *
 * @property {UUIDString} [applicationId]
 * @property {number} [end]
 * @property {number} [start]
 * @property {UUIDString} [userId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} LoginRecordSearchRequest
 *
 * @property {boolean} [retrieveTotal]
 * @property {LoginRecordSearchCriteria} [search]
 */


/**
 * A raw login record response
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} LoginRecordSearchResponse
 *
 * @property {Array<DisplayableRawLogin>} [logins]
 * @property {number} [total]
 */


/**
 * Response for the login report.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} LoginReportResponse
 *
 * @property {Array<Count>} [hourlyCounts]
 * @property {number} [total]
 */


/**
 * Login API request object.
 *
 * @author Seth Musselman
 *
 * @typedef {Object} LoginRequest
 * @extends BaseLoginRequest
 *
 * @property {string} [loginId]
 * @property {string} [oneTimePassword]
 * @property {string} [password]
 * @property {string} [twoFactorTrustId]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} LoginResponse
 *
 * @property {Array<LoginPreventedResponse>} [actions]
 * @property {string} [changePasswordId]
 * @property {ChangePasswordReason} [changePasswordReason]
 * @property {string} [emailVerificationId]
 * @property {Array<TwoFactorMethod>} [methods]
 * @property {string} [pendingIdPLinkId]
 * @property {string} [refreshToken]
 * @property {string} [registrationVerificationId]
 * @property {Object<string, Object>} [state]
 * @property {string} [token]
 * @property {string} [twoFactorId]
 * @property {string} [twoFactorTrustId]
 * @property {User} [user]
 */


/**
 * @author Matthew Altman
 *
 * @readonly
 * @enum
 */
var LogoutBehavior = {
  RedirectOnly: "RedirectOnly",
  AllApplications: "AllApplications"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} LookupResponse
 *
 * @property {IdentityProviderDetails} [identityProvider]
 */


/**
 * This class contains the managed fields that are also put into the database during FusionAuth setup.
 * <p>
 * NOTE TO FUSIONAUTH DEVS: These fields are are also declared in SQL in order to boot strap the system. These need to stay in sync.
 * - Any changes to these fields needs to also be reflected in mysql.sql and postgresql.sql
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ManagedFields
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} MaximumPasswordAge
 * @extends Enableable
 *
 * @property {number} [days]
 */


/**
 * Group Member Delete Request
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} MemberDeleteRequest
 *
 * @property {Array<UUIDString>} [memberIds]
 * @property {Object<UUIDString, Array<UUIDString>>} [members]
 */


/**
 * Group Member Request
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} MemberRequest
 *
 * @property {Object<UUIDString, Array<GroupMember>>} [members]
 */


/**
 * Group Member Response
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} MemberResponse
 *
 * @property {Object<UUIDString, Array<GroupMember>>} [members]
 */


/**
 * @author Mikey Sleevi
 *
 * @typedef {Object} Message
 *
 */


/**
 * Stores an message template used to distribute messages;
 *
 * @author Michael Sleevi
 *
 * @typedef {Object} MessageTemplate
 *
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {string} [name]
 * @property {MessageType} [type]
 */


/**
 * A Message Template Request to the API
 *
 * @author Michael Sleevi
 *
 * @typedef {Object} MessageTemplateRequest
 *
 * @property {MessageTemplate} [messageTemplate]
 */


/**
 * @author Michael Sleevi
 *
 * @typedef {Object} MessageTemplateResponse
 *
 * @property {MessageTemplate} [messageTemplate]
 * @property {Array<MessageTemplate>} [messageTemplates]
 */


/**
 * @author Mikey Sleevi
 *
 * @readonly
 * @enum
 */
var MessageType = {
  SMS: "SMS"
};

/**
 * @author Brett Guy
 *
 * @typedef {Object} MessengerRequest
 *
 * @property {BaseMessengerConfiguration} [messenger]
 */


/**
 * @author Brett Guy
 *
 * @typedef {Object} MessengerResponse
 *
 * @property {BaseMessengerConfiguration} [messenger]
 * @property {Array<BaseMessengerConfiguration>} [messengers]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} MessengerTransport
 *
 */


/**
 * @author Brett Guy
 *
 * @readonly
 * @enum
 */
var MessengerType = {
  Generic: "Generic",
  Kafka: "Kafka",
  Twilio: "Twilio"
};

/**
 * @typedef {Object} MetaData
 *
 * @property {DeviceInfo} [device]
 * @property {Object<string>} [scopes]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} MinimumPasswordAge
 * @extends Enableable
 *
 * @property {number} [seconds]
 */


/**
 * Response for the daily active user report.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} MonthlyActiveUserReportResponse
 *
 * @property {Array<Count>} [monthlyActiveUsers]
 * @property {number} [total]
 */


/**
 * @typedef {Object} MultiFactorAuthenticatorMethod
 * @extends Enableable
 *
 * @property {TOTPAlgorithm} [algorithm]
 * @property {number} [codeLength]
 * @property {number} [timeStep]
 */


/**
 * @typedef {Object} MultiFactorEmailMethod
 * @extends Enableable
 *
 * @property {UUIDString} [templateId]
 */


/**
 * @typedef {Object} MultiFactorEmailTemplate
 *
 * @property {UUIDString} [templateId]
 */


/**
 * @typedef {Object} MultiFactorSMSMethod
 * @extends Enableable
 *
 * @property {UUIDString} [messengerId]
 * @property {UUIDString} [templateId]
 */


/**
 * @typedef {Object} MultiFactorSMSTemplate
 *
 * @property {UUIDString} [templateId]
 */


/**
 * @author Brett Pontarelli
 *
 * @typedef {Object} NintendoApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Nintendo gaming login provider.
 *
 * @author Brett Pontarelli
 *
 * @typedef {Object} NintendoIdentityProvider
 * @extends BaseIdentityProvider<NintendoApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Helper methods for normalizing values.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Normalizer
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} OAuth2Configuration
 *
 * @property {Array<string>} [authorizedOriginURLs]
 * @property {Array<string>} [authorizedRedirectURLs]
 * @property {ClientAuthenticationPolicy} [clientAuthenticationPolicy]
 * @property {string} [clientId]
 * @property {string} [clientSecret]
 * @property {boolean} [debug]
 * @property {string} [deviceVerificationURL]
 * @property {Object<GrantType>} [enabledGrants]
 * @property {boolean} [generateRefreshTokens]
 * @property {LogoutBehavior} [logoutBehavior]
 * @property {string} [logoutURL]
 * @property {ProofKeyForCodeExchangePolicy} [proofKeyForCodeExchangePolicy]
 * @property {boolean} [requireClientAuthentication]
 * @property {boolean} [requireRegistration]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} OAuthConfigurationResponse
 *
 * @property {number} [httpSessionMaxInactiveInterval]
 * @property {string} [logoutURL]
 * @property {OAuth2Configuration} [oauthConfiguration]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} OAuthError
 *
 * @property {string} [change_password_id]
 * @property {OAuthErrorType} [error]
 * @property {string} [error_description]
 * @property {OAuthErrorReason} [error_reason]
 * @property {string} [error_uri]
 * @property {string} [two_factor_id]
 */


/**
 * @readonly
 * @enum
 */
var OAuthErrorReason = {
  auth_code_not_found: "auth_code_not_found",
  access_token_malformed: "access_token_malformed",
  access_token_expired: "access_token_expired",
  access_token_unavailable_for_processing: "access_token_unavailable_for_processing",
  access_token_failed_processing: "access_token_failed_processing",
  refresh_token_not_found: "refresh_token_not_found",
  refresh_token_type_not_supported: "refresh_token_type_not_supported",
  invalid_client_id: "invalid_client_id",
  invalid_user_credentials: "invalid_user_credentials",
  invalid_grant_type: "invalid_grant_type",
  invalid_origin: "invalid_origin",
  invalid_origin_opaque: "invalid_origin_opaque",
  invalid_pkce_code_verifier: "invalid_pkce_code_verifier",
  invalid_pkce_code_challenge: "invalid_pkce_code_challenge",
  invalid_pkce_code_challenge_method: "invalid_pkce_code_challenge_method",
  invalid_redirect_uri: "invalid_redirect_uri",
  invalid_response_mode: "invalid_response_mode",
  invalid_response_type: "invalid_response_type",
  invalid_id_token_hint: "invalid_id_token_hint",
  invalid_post_logout_redirect_uri: "invalid_post_logout_redirect_uri",
  invalid_device_code: "invalid_device_code",
  invalid_user_code: "invalid_user_code",
  invalid_additional_client_id: "invalid_additional_client_id",
  invalid_target_entity_scope: "invalid_target_entity_scope",
  invalid_entity_permission_scope: "invalid_entity_permission_scope",
  grant_type_disabled: "grant_type_disabled",
  missing_client_id: "missing_client_id",
  missing_client_secret: "missing_client_secret",
  missing_code: "missing_code",
  missing_code_challenge: "missing_code_challenge",
  missing_code_verifier: "missing_code_verifier",
  missing_device_code: "missing_device_code",
  missing_grant_type: "missing_grant_type",
  missing_redirect_uri: "missing_redirect_uri",
  missing_refresh_token: "missing_refresh_token",
  missing_response_type: "missing_response_type",
  missing_token: "missing_token",
  missing_user_code: "missing_user_code",
  missing_verification_uri: "missing_verification_uri",
  login_prevented: "login_prevented",
  not_licensed: "not_licensed",
  user_code_expired: "user_code_expired",
  user_expired: "user_expired",
  user_locked: "user_locked",
  user_not_found: "user_not_found",
  client_authentication_missing: "client_authentication_missing",
  invalid_client_authentication_scheme: "invalid_client_authentication_scheme",
  invalid_client_authentication: "invalid_client_authentication",
  client_id_mismatch: "client_id_mismatch",
  change_password_administrative: "change_password_administrative",
  change_password_breached: "change_password_breached",
  change_password_expired: "change_password_expired",
  change_password_validation: "change_password_validation",
  unknown: "unknown"
};

/**
 * @readonly
 * @enum
 */
var OAuthErrorType = {
  invalid_request: "invalid_request",
  invalid_client: "invalid_client",
  invalid_grant: "invalid_grant",
  invalid_token: "invalid_token",
  unauthorized_client: "unauthorized_client",
  invalid_scope: "invalid_scope",
  server_error: "server_error",
  unsupported_grant_type: "unsupported_grant_type",
  unsupported_response_type: "unsupported_response_type",
  change_password_required: "change_password_required",
  not_licensed: "not_licensed",
  two_factor_required: "two_factor_required",
  authorization_pending: "authorization_pending",
  expired_token: "expired_token",
  unsupported_token_type: "unsupported_token_type"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} OAuthResponse
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var ObjectState = {
  Active: "Active",
  Inactive: "Inactive",
  PendingDelete: "PendingDelete"
};

/**
 * OpenID Connect Configuration as described by the <a href="https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderMetadata">OpenID
 * Provider Metadata</a>.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} OpenIdConfiguration
 *
 * @property {string} [authorization_endpoint]
 * @property {boolean} [backchannel_logout_supported]
 * @property {Array<string>} [claims_supported]
 * @property {string} [device_authorization_endpoint]
 * @property {string} [end_session_endpoint]
 * @property {boolean} [frontchannel_logout_supported]
 * @property {Array<string>} [grant_types_supported]
 * @property {Array<string>} [id_token_signing_alg_values_supported]
 * @property {string} [issuer]
 * @property {string} [jwks_uri]
 * @property {Array<string>} [response_modes_supported]
 * @property {Array<string>} [response_types_supported]
 * @property {Array<string>} [scopes_supported]
 * @property {Array<string>} [subject_types_supported]
 * @property {string} [token_endpoint]
 * @property {Array<string>} [token_endpoint_auth_methods_supported]
 * @property {string} [userinfo_endpoint]
 * @property {Array<string>} [userinfo_signing_alg_values_supported]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} OpenIdConnectApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonImageURL]
 * @property {string} [buttonText]
 * @property {IdentityProviderOauth2Configuration} [oauth2]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} OpenIdConnectIdentityProvider
 * @extends BaseIdentityProvider<OpenIdConnectApplicationConfiguration>
 *
 * @property {string} [buttonImageURL]
 * @property {string} [buttonText]
 * @property {Object<string>} [domains]
 * @property {IdentityProviderOauth2Configuration} [oauth2]
 * @property {boolean} [postRequest]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordBreachDetection
 * @extends Enableable
 *
 * @property {BreachMatchMode} [matchMode]
 * @property {UUIDString} [notifyUserEmailTemplateId]
 * @property {BreachAction} [onLogin]
 */


/**
 * Password Encryption Scheme Configuration
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordEncryptionConfiguration
 *
 * @property {string} [encryptionScheme]
 * @property {number} [encryptionSchemeFactor]
 * @property {boolean} [modifyEncryptionSchemeOnLogin]
 */


/**
 * @typedef {Object} PasswordlessConfiguration
 * @extends Enableable
 *
 */


/**
 * Interface for all identity providers that are passwordless and do not accept a password.
 *
 * @typedef {Object} PasswordlessIdentityProvider
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordlessLoginRequest
 * @extends BaseLoginRequest
 *
 * @property {string} [code]
 * @property {string} [twoFactorTrustId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordlessSendRequest
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [code]
 * @property {string} [loginId]
 * @property {Object<string, Object>} [state]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordlessStartRequest
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [loginId]
 * @property {Object<string, Object>} [state]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordlessStartResponse
 *
 * @property {string} [code]
 */


/**
 * @author Derek Klatt
 *
 * @typedef {Object} PasswordValidationRules
 *
 * @property {PasswordBreachDetection} [breachDetection]
 * @property {number} [maxLength]
 * @property {number} [minLength]
 * @property {RememberPreviousPasswords} [rememberPreviousPasswords]
 * @property {boolean} [requireMixedCase]
 * @property {boolean} [requireNonAlpha]
 * @property {boolean} [requireNumber]
 * @property {boolean} [validateOnLogin]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PasswordValidationRulesResponse
 *
 * @property {PasswordValidationRules} [passwordValidationRules]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} PendingIdPLink
 *
 * @property {string} [displayName]
 * @property {string} [email]
 * @property {UUIDString} [identityProviderId]
 * @property {string} [identityProviderName]
 * @property {IdentityProviderType} [identityProviderType]
 * @property {string} [identityProviderUserId]
 * @property {User} [user]
 * @property {string} [username]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} PendingResponse
 *
 * @property {Array<User>} [users]
 */


/**
 * @author Michael Sleevi
 *
 * @typedef {Object} PreviewMessageTemplateRequest
 *
 * @property {string} [locale]
 * @property {MessageTemplate} [messageTemplate]
 */


/**
 * @author Michael Sleevi
 *
 * @typedef {Object} PreviewMessageTemplateResponse
 *
 * @property {Errors} [errors]
 * @property {SMSMessage} [message]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} PreviewRequest
 *
 * @property {EmailTemplate} [emailTemplate]
 * @property {string} [locale]
 */


/**
 * @author Seth Musselman
 *
 * @typedef {Object} PreviewResponse
 *
 * @property {Email} [email]
 * @property {Errors} [errors]
 */


/**
 * @author Brett Guy
 *
 * @readonly
 * @enum
 */
var ProofKeyForCodeExchangePolicy = {
  Required: "Required",
  NotRequired: "NotRequired",
  NotRequiredWhenUsingClientAuthentication: "NotRequiredWhenUsingClientAuthentication"
};

/**
 * JWT Public Key Response Object
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} PublicKeyResponse
 *
 * @property {string} [publicKey]
 * @property {Object<string, string>} [publicKeys]
 */


/**
 * Raw login information for each time a user logs into an application.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} RawLogin
 *
 * @property {UUIDString} [applicationId]
 * @property {number} [instant]
 * @property {string} [ipAddress]
 * @property {UUIDString} [userId]
 */


/**
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var ReactorFeatureStatus = {
  ACTIVE: "ACTIVE",
  DISCONNECTED: "DISCONNECTED",
  PENDING: "PENDING",
  UNKNOWN: "UNKNOWN"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ReactorMetrics
 *
 * @property {Object<UUIDString, BreachedPasswordTenantMetric>} [breachedPasswordMetrics]
 */


/**
 * Request for managing FusionAuth Reactor and licenses.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} ReactorRequest
 *
 * @property {string} [license]
 * @property {string} [licenseId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ReactorResponse
 *
 * @property {ReactorMetrics} [metrics]
 * @property {ReactorStatus} [status]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ReactorStatus
 *
 * @property {ReactorFeatureStatus} [advancedIdentityProviders]
 * @property {ReactorFeatureStatus} [advancedMultiFactorAuthentication]
 * @property {ReactorFeatureStatus} [advancedRegistration]
 * @property {ReactorFeatureStatus} [applicationThemes]
 * @property {ReactorFeatureStatus} [breachedPasswordDetection]
 * @property {ReactorFeatureStatus} [connectors]
 * @property {ReactorFeatureStatus} [entityManagement]
 * @property {boolean} [licensed]
 */


/**
 * Response for the user login report.
 *
 * @author Seth Musselman
 *
 * @typedef {Object} RecentLoginResponse
 *
 * @property {Array<DisplayableRawLogin>} [logins]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} RefreshRequest
 *
 * @property {string} [refreshToken]
 * @property {string} [token]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} RefreshResponse
 *
 */


/**
 * Models a JWT Refresh Token.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} RefreshToken
 *
 * @property {UUIDString} [applicationId]
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {MetaData} [metaData]
 * @property {number} [startInstant]
 * @property {string} [token]
 * @property {UUIDString} [userId]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var RefreshTokenExpirationPolicy = {
  Fixed: "Fixed",
  SlidingWindow: "SlidingWindow"
};

/**
 * Refresh Token Import request.
 *
 * @author Brett Guy
 *
 * @typedef {Object} RefreshTokenImportRequest
 *
 * @property {Array<RefreshToken>} [refreshTokens]
 * @property {boolean} [validateDbConstraints]
 */


/**
 * API response for retrieving Refresh Tokens
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} RefreshTokenResponse
 *
 * @property {RefreshToken} [refreshToken]
 * @property {Array<RefreshToken>} [refreshTokens]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} RefreshTokenRevocationPolicy
 *
 * @property {boolean} [onLoginPrevented]
 * @property {boolean} [onPasswordChanged]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var RefreshTokenUsagePolicy = {
  Reusable: "Reusable",
  OneTimeUse: "OneTimeUse"
};

/**
 * @typedef {Object} RegistrationConfiguration
 * @extends Enableable
 *
 * @property {Requirable} [birthDate]
 * @property {boolean} [confirmPassword]
 * @property {Requirable} [firstName]
 * @property {UUIDString} [formId]
 * @property {Requirable} [fullName]
 * @property {Requirable} [lastName]
 * @property {LoginIdType} [loginIdType]
 * @property {Requirable} [middleName]
 * @property {Requirable} [mobilePhone]
 * @property {RegistrationType} [type]
 */


/**
 * Response for the registration report.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} RegistrationReportResponse
 *
 * @property {Array<Count>} [hourlyCounts]
 * @property {number} [total]
 */


/**
 * Registration API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} RegistrationRequest
 *
 * @property {boolean} [generateAuthenticationToken]
 * @property {UserRegistration} [registration]
 * @property {boolean} [sendSetPasswordEmail]
 * @property {boolean} [skipRegistrationVerification]
 * @property {boolean} [skipVerification]
 * @property {User} [user]
 */


/**
 * Registration API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} RegistrationResponse
 *
 * @property {string} [refreshToken]
 * @property {UserRegistration} [registration]
 * @property {string} [registrationVerificationId]
 * @property {string} [token]
 * @property {User} [user]
 */


/**
 * @readonly
 * @enum
 */
var RegistrationType = {
  basic: "basic",
  advanced: "advanced"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} RegistrationUnverifiedOptions
 *
 * @property {UnverifiedBehavior} [behavior]
 */


/**
 * Reindex API request
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} ReindexRequest
 *
 * @property {string} [index]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ReloadRequest
 *
 * @property {Array<string>} [names]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} RememberPreviousPasswords
 * @extends Enableable
 *
 * @property {number} [count]
 */


/**
 * Something that can be required and thus also optional. This currently extends Enableable because anything that is
 * require/optional is almost always enableable as well.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Requirable
 * @extends Enableable
 *
 * @property {boolean} [required]
 */


/**
 * Interface describing the need for CORS configuration.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} RequiresCORSConfiguration
 *
 */


/**
 * @readonly
 * @enum
 */
var SAMLLogoutBehavior = {
  AllParticipants: "AllParticipants",
  OnlyOriginator: "OnlyOriginator"
};

/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} SAMLv2ApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonImageURL]
 * @property {string} [buttonText]
 */


/**
 * @typedef {Object} SAMLv2Configuration
 * @extends Enableable
 *
 * @property {string} [audience]
 * @property {Array<string>} [authorizedRedirectURLs]
 * @property {string} [callbackURL]
 * @property {boolean} [debug]
 * @property {UUIDString} [defaultVerificationKeyId]
 * @property {string} [issuer]
 * @property {UUIDString} [keyId]
 * @property {SAMLv2Logout} [logout]
 * @property {string} [logoutURL]
 * @property {boolean} [requireSignedRequests]
 * @property {CanonicalizationMethod} [xmlSignatureC14nMethod]
 * @property {XMLSignatureLocation} [xmlSignatureLocation]
 */


/**
 * SAML v2 identity provider configuration.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SAMLv2IdentityProvider
 * @extends BaseIdentityProvider<SAMLv2ApplicationConfiguration>
 *
 * @property {string} [buttonImageURL]
 * @property {string} [buttonText]
 * @property {Object<string>} [domains]
 * @property {string} [emailClaim]
 * @property {string} [idpEndpoint]
 * @property {string} [issuer]
 * @property {UUIDString} [keyId]
 * @property {string} [nameIdFormat]
 * @property {boolean} [postRequest]
 * @property {UUIDString} [requestSigningKeyId]
 * @property {boolean} [signRequest]
 * @property {string} [uniqueIdClaim]
 * @property {boolean} [useNameIdForEmail]
 * @property {string} [usernameClaim]
 * @property {CanonicalizationMethod} [xmlSignatureC14nMethod]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SAMLv2IdPInitiatedApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 */


/**
 * SAML v2 IdP Initiated identity provider configuration.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} SAMLv2IdPInitiatedIdentityProvider
 * @extends BaseIdentityProvider<SAMLv2IdPInitiatedApplicationConfiguration>
 *
 * @property {string} [emailClaim]
 * @property {string} [issuer]
 * @property {UUIDString} [keyId]
 * @property {string} [uniqueIdClaim]
 * @property {boolean} [useNameIdForEmail]
 * @property {string} [usernameClaim]
 */


/**
 * @typedef {Object} SAMLv2Logout
 *
 * @property {SAMLLogoutBehavior} [behavior]
 * @property {UUIDString} [defaultVerificationKeyId]
 * @property {UUIDString} [keyId]
 * @property {boolean} [requireSignedRequests]
 * @property {SAMLv2SingleLogout} [singleLogout]
 * @property {CanonicalizationMethod} [xmlSignatureC14nMethod]
 */


/**
 * @typedef {Object} SAMLv2SingleLogout
 * @extends Enableable
 *
 * @property {UUIDString} [keyId]
 * @property {string} [url]
 * @property {CanonicalizationMethod} [xmlSignatureC14nMethod]
 */


/**
 * Search API request.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SearchRequest
 *
 * @property {UserSearchCriteria} [search]
 */


/**
 * Search API response.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SearchResponse
 *
 * @property {number} [total]
 * @property {Array<User>} [users]
 */


/**
 * Search results.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SearchResults
 * @template {T}
 *
 * @property {Array<T>} [results]
 * @property {number} [total]
 * @property {boolean} [totalEqualToActual]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SecretResponse
 *
 * @property {string} [secret]
 * @property {string} [secretBase32Encoded]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SecureGeneratorConfiguration
 *
 * @property {number} [length]
 * @property {SecureGeneratorType} [type]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var SecureGeneratorType = {
  randomDigits: "randomDigits",
  randomBytes: "randomBytes",
  randomAlpha: "randomAlpha",
  randomAlphaNumeric: "randomAlphaNumeric"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SecureIdentity
 *
 * @property {number} [breachedPasswordLastCheckedInstant]
 * @property {BreachedPasswordStatus} [breachedPasswordStatus]
 * @property {UUIDString} [connectorId]
 * @property {string} [encryptionScheme]
 * @property {number} [factor]
 * @property {UUIDString} [id]
 * @property {number} [lastLoginInstant]
 * @property {string} [password]
 * @property {ChangePasswordReason} [passwordChangeReason]
 * @property {boolean} [passwordChangeRequired]
 * @property {number} [passwordLastUpdateInstant]
 * @property {string} [salt]
 * @property {string} [uniqueUsername]
 * @property {string} [username]
 * @property {ContentStatus} [usernameStatus]
 * @property {boolean} [verified]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SendRequest
 *
 * @property {UUIDString} [applicationId]
 * @property {Array<string>} [bccAddresses]
 * @property {Array<string>} [ccAddresses]
 * @property {Array<string>} [preferredLanguages]
 * @property {Object<string, Object>} [requestData]
 * @property {Array<EmailAddress>} [toAddresses]
 * @property {Array<UUIDString>} [userIds]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SendResponse
 *
 * @property {Object<string, EmailTemplateErrors>} [anonymousResults]
 * @property {Object<UUIDString, EmailTemplateErrors>} [results]
 */


/**
 * @author Michael Sleevi
 *
 * @typedef {Object} SMSMessage
 *
 * @property {string} [phoneNumber]
 * @property {string} [textMessage]
 */


/**
 * @author Michael Sleevi
 *
 * @typedef {Object} SMSMessageTemplate
 * @extends MessageTemplate
 *
 * @property {string} [defaultTemplate]
 * @property {LocalizedStrings} [localizedTemplates]
 */


/**
 * @author Brett Pontarelli
 *
 * @typedef {Object} SonyPSNApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * SonyPSN gaming login provider.
 *
 * @author Brett Pontarelli
 *
 * @typedef {Object} SonyPSNIdentityProvider
 * @extends BaseIdentityProvider<SonyPSNApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var Sort = {
  asc: "asc",
  desc: "desc"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SortField
 *
 * @property {string} [missing]
 * @property {string} [name]
 * @property {Sort} [order]
 */


/**
 * @author Brett Pontarelli
 *
 * @typedef {Object} SteamApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [scope]
 * @property {string} [webAPIKey]
 */


/**
 * Steam gaming login provider.
 *
 * @author Brett Pontarelli
 *
 * @typedef {Object} SteamIdentityProvider
 * @extends BaseIdentityProvider<SteamApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [scope]
 * @property {string} [webAPIKey]
 */


/**
 * Helper interface that indicates an identity provider can be federated to using the HTTP POST method.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SupportsPostBindings
 *
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} SystemConfiguration
 *
 * @property {AuditLogConfiguration} [auditLogConfiguration]
 * @property {CORSConfiguration} [corsConfiguration]
 * @property {Object<string, Object>} [data]
 * @property {EventLogConfiguration} [eventLogConfiguration]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {LoginRecordConfiguration} [loginRecordConfiguration]
 * @property {string} [reportTimezone]
 * @property {UIConfiguration} [uiConfiguration]
 */


/**
 * Request for the system configuration API.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SystemConfigurationRequest
 *
 * @property {SystemConfiguration} [systemConfiguration]
 */


/**
 * Response for the system configuration API.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} SystemConfigurationResponse
 *
 * @property {SystemConfiguration} [systemConfiguration]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} SystemLogsExportRequest
 * @extends BaseExportRequest
 *
 * @property {number} [lastNBytes]
 */


/**
 * @typedef {Object} Templates
 *
 * @property {string} [accountEdit]
 * @property {string} [accountIndex]
 * @property {string} [accountTwoFactorDisable]
 * @property {string} [accountTwoFactorEnable]
 * @property {string} [accountTwoFactorIndex]
 * @property {string} [emailComplete]
 * @property {string} [emailSend]
 * @property {string} [emailSent]
 * @property {string} [emailVerificationRequired]
 * @property {string} [emailVerify]
 * @property {string} [helpers]
 * @property {string} [index]
 * @property {string} [oauth2Authorize]
 * @property {string} [oauth2AuthorizedNotRegistered]
 * @property {string} [oauth2ChildRegistrationNotAllowed]
 * @property {string} [oauth2ChildRegistrationNotAllowedComplete]
 * @property {string} [oauth2CompleteRegistration]
 * @property {string} [oauth2Device]
 * @property {string} [oauth2DeviceComplete]
 * @property {string} [oauth2Error]
 * @property {string} [oauth2Logout]
 * @property {string} [oauth2Passwordless]
 * @property {string} [oauth2Register]
 * @property {string} [oauth2StartIdPLink]
 * @property {string} [oauth2TwoFactor]
 * @property {string} [oauth2TwoFactorMethods]
 * @property {string} [oauth2Wait]
 * @property {string} [passwordChange]
 * @property {string} [passwordComplete]
 * @property {string} [passwordForgot]
 * @property {string} [passwordSent]
 * @property {string} [registrationComplete]
 * @property {string} [registrationSend]
 * @property {string} [registrationSent]
 * @property {string} [registrationVerificationRequired]
 * @property {string} [registrationVerify]
 * @property {string} [samlv2Logout]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} Tenant
 *
 * @property {boolean} [configured]
 * @property {Array<ConnectorPolicy>} [connectorPolicies]
 * @property {Object<string, Object>} [data]
 * @property {EmailConfiguration} [emailConfiguration]
 * @property {EventConfiguration} [eventConfiguration]
 * @property {ExternalIdentifierConfiguration} [externalIdentifierConfiguration]
 * @property {FailedAuthenticationConfiguration} [failedAuthenticationConfiguration]
 * @property {FamilyConfiguration} [familyConfiguration]
 * @property {TenantFormConfiguration} [formConfiguration]
 * @property {number} [httpSessionMaxInactiveInterval]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {string} [issuer]
 * @property {JWTConfiguration} [jwtConfiguration]
 * @property {number} [lastUpdateInstant]
 * @property {TenantLoginConfiguration} [loginConfiguration]
 * @property {string} [logoutURL]
 * @property {MaximumPasswordAge} [maximumPasswordAge]
 * @property {MinimumPasswordAge} [minimumPasswordAge]
 * @property {TenantMultiFactorConfiguration} [multiFactorConfiguration]
 * @property {string} [name]
 * @property {TenantOAuth2Configuration} [oauthConfiguration]
 * @property {PasswordEncryptionConfiguration} [passwordEncryptionConfiguration]
 * @property {PasswordValidationRules} [passwordValidationRules]
 * @property {ObjectState} [state]
 * @property {UUIDString} [themeId]
 * @property {TenantUserDeletePolicy} [userDeletePolicy]
 * @property {TenantUsernameConfiguration} [usernameConfiguration]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} Tenantable
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TenantFormConfiguration
 *
 * @property {UUIDString} [adminUserFormId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TenantLoginConfiguration
 *
 * @property {boolean} [requireAuthentication]
 */


/**
 * @author Mikey Sleevi
 *
 * @typedef {Object} TenantMultiFactorConfiguration
 *
 * @property {MultiFactorAuthenticatorMethod} [authenticator]
 * @property {MultiFactorEmailMethod} [email]
 * @property {MultiFactorSMSMethod} [sms]
 */


/**
 * @typedef {Object} TenantOAuth2Configuration
 *
 * @property {UUIDString} [clientCredentialsAccessTokenPopulateLambdaId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TenantRequest
 *
 * @property {UUIDString} [sourceTenantId]
 * @property {Tenant} [tenant]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TenantResponse
 *
 * @property {Tenant} [tenant]
 * @property {Array<Tenant>} [tenants]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TenantUnverifiedConfiguration
 *
 * @property {UnverifiedBehavior} [email]
 * @property {RegistrationUnverifiedOptions} [whenGated]
 */


/**
 * A Tenant-level policy for deleting Users.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} TenantUserDeletePolicy
 *
 * @property {TimeBasedDeletePolicy} [unverified]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TenantUsernameConfiguration
 *
 * @property {UniqueUsernameConfiguration} [unique]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TestEvent
 * @extends BaseEvent
 *
 * @property {string} [message]
 */


/**
 * @author Trevor Smith
 *
 * @typedef {Object} Theme
 *
 * @property {Object<string, Object>} [data]
 * @property {string} [defaultMessages]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {LocalizedStrings} [localizedMessages]
 * @property {string} [name]
 * @property {string} [stylesheet]
 * @property {Templates} [templates]
 */


/**
 * Theme API request object.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} ThemeRequest
 *
 * @property {UUIDString} [sourceThemeId]
 * @property {Theme} [theme]
 */


/**
 * Theme API response object.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} ThemeResponse
 *
 * @property {Theme} [theme]
 * @property {Array<Theme>} [themes]
 */


/**
 * A policy for deleting Users.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} TimeBasedDeletePolicy
 * @extends Enableable
 *
 * @property {number} [numberOfDaysToRetain]
 */


/**
 * <ul>
 * <li>Bearer Token type as defined by <a href="https://tools.ietf.org/html/rfc6750">RFC 6750</a>.</li>
 * <li>MAC Token type as referenced by <a href="https://tools.ietf.org/html/rfc6749">RFC 6749</a> and
 * <a href="https://tools.ietf.org/html/draft-ietf-oauth-v2-http-mac-05">
 * Draft RFC on OAuth 2.0 Message Authentication Code (MAC) Tokens</a>
 * </li>
 * </ul>
 *
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var TokenType = {
  Bearer: "Bearer",
  MAC: "MAC"
};

/**
 * @typedef {Object} Totals
 *
 * @property {number} [logins]
 * @property {number} [registrations]
 * @property {number} [totalRegistrations]
 */


/**
 * The response from the total report. This report stores the total numbers for each application.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} TotalsReportResponse
 *
 * @property {Object<UUIDString, Totals>} [applicationTotals]
 * @property {number} [globalRegistrations]
 * @property {number} [totalGlobalRegistrations]
 */


/**
 * @readonly
 * @enum
 */
var TOTPAlgorithm = {
  HmacSHA1: "HmacSHA1",
  HmacSHA256: "HmacSHA256",
  HmacSHA512: "HmacSHA512"
};

/**
 * The transaction types for Webhooks and other event systems within FusionAuth.
 *
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var TransactionType = {
  None: "None",
  Any: "Any",
  SimpleMajority: "SimpleMajority",
  SuperMajority: "SuperMajority",
  AbsoluteMajority: "AbsoluteMajority"
};

/**
 * @author Brett Guy
 *
 * @typedef {Object} TwilioMessengerConfiguration
 * @extends BaseMessengerConfiguration
 *
 * @property {string} [accountSID]
 * @property {string} [authToken]
 * @property {string} [fromPhoneNumber]
 * @property {string} [messagingServiceSid]
 * @property {string} [url]
 */


/**
 * @author Brett Pontarelli
 *
 * @typedef {Object} TwitchApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Twitch gaming login provider.
 *
 * @author Brett Pontarelli
 *
 * @typedef {Object} TwitchIdentityProvider
 * @extends BaseIdentityProvider<TwitchApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwitterApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [consumerKey]
 * @property {string} [consumerSecret]
 */


/**
 * Twitter social login provider.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwitterIdentityProvider
 * @extends BaseIdentityProvider<TwitterApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [consumerKey]
 * @property {string} [consumerSecret]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorEnableDisableSendRequest
 *
 * @property {string} [email]
 * @property {string} [method]
 * @property {string} [methodId]
 * @property {string} [mobilePhone]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorLoginRequest
 * @extends BaseLoginRequest
 *
 * @property {string} [code]
 * @property {boolean} [trustComputer]
 * @property {string} [twoFactorId]
 * @property {UUIDString} [userId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorMethod
 *
 * @property {AuthenticatorConfiguration} [authenticator]
 * @property {string} [email]
 * @property {string} [id]
 * @property {boolean} [lastUsed]
 * @property {string} [method]
 * @property {string} [mobilePhone]
 * @property {string} [secret]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorRecoveryCodeResponse
 *
 * @property {Array<string>} [recoveryCodes]
 */


/**
 * @author Brian Pontarelli
 *
 * @typedef {Object} TwoFactorRequest
 *
 * @property {string} [authenticatorId]
 * @property {string} [code]
 * @property {string} [email]
 * @property {string} [method]
 * @property {string} [mobilePhone]
 * @property {string} [secret]
 * @property {string} [secretBase32Encoded]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorResponse
 *
 * @property {Array<string>} [recoveryCodes]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorSendRequest
 *
 * @property {string} [email]
 * @property {string} [method]
 * @property {string} [methodId]
 * @property {string} [mobilePhone]
 * @property {UUIDString} [userId]
 */


/**
 * @author Brett Guy
 *
 * @typedef {Object} TwoFactorStartRequest
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [code]
 * @property {string} [loginId]
 * @property {Object<string, Object>} [state]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} TwoFactorStartResponse
 *
 * @property {string} [code]
 * @property {Array<TwoFactorMethod>} [methods]
 * @property {string} [twoFactorId]
 */


/**
 * @typedef {Object} UIConfiguration
 *
 * @property {string} [headerColor]
 * @property {string} [logoURL]
 * @property {string} [menuFontColor]
 */


/**
 * @typedef {Object} UniqueUsernameConfiguration
 * @extends Enableable
 *
 * @property {number} [numberOfDigits]
 * @property {char} [separator]
 * @property {UniqueUsernameStrategy} [strategy]
 */


/**
 * @readonly
 * @enum
 */
var UniqueUsernameStrategy = {
  Always: "Always",
  OnCollision: "OnCollision"
};

/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var UnverifiedBehavior = {
  Allow: "Allow",
  Gated: "Gated"
};

/**
 * The global view of a User. This object contains all global information about the user including birth date, registration information
 * preferred languages, global attributes, etc.
 *
 * @author Seth Musselman
 *
 * @typedef {Object} User
 * @extends SecureIdentity
 *
 * @property {boolean} [active]
 * @property {string} [birthDate]
 * @property {UUIDString} [cleanSpeakId]
 * @property {Object<string, Object>} [data]
 * @property {string} [email]
 * @property {number} [expiry]
 * @property {string} [firstName]
 * @property {string} [fullName]
 * @property {string} [imageUrl]
 * @property {number} [insertInstant]
 * @property {string} [lastName]
 * @property {number} [lastUpdateInstant]
 * @property {Array<GroupMember>} [memberships]
 * @property {string} [middleName]
 * @property {string} [mobilePhone]
 * @property {string} [parentEmail]
 * @property {Array<string>} [preferredLanguages]
 * @property {Array<UserRegistration>} [registrations]
 * @property {UUIDString} [tenantId]
 * @property {string} [timezone]
 * @property {UserTwoFactorConfiguration} [twoFactor]
 */


/**
 * An action that can be executed on a user (discipline or reward potentially).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserAction
 *
 * @property {boolean} [active]
 * @property {UUIDString} [cancelEmailTemplateId]
 * @property {UUIDString} [endEmailTemplateId]
 * @property {UUIDString} [id]
 * @property {boolean} [includeEmailInEventJSON]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {LocalizedStrings} [localizedNames]
 * @property {UUIDString} [modifyEmailTemplateId]
 * @property {string} [name]
 * @property {Array<UserActionOption>} [options]
 * @property {boolean} [preventLogin]
 * @property {boolean} [sendEndEvent]
 * @property {UUIDString} [startEmailTemplateId]
 * @property {boolean} [temporal]
 * @property {TransactionType} [transactionType]
 * @property {boolean} [userEmailingEnabled]
 * @property {boolean} [userNotificationsEnabled]
 */


/**
 * Models the user action event (and can be converted to JSON).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionEvent
 * @extends BaseEvent
 *
 * @property {string} [action]
 * @property {UUIDString} [actioneeUserId]
 * @property {UUIDString} [actionerUserId]
 * @property {UUIDString} [actionId]
 * @property {Array<UUIDString>} [applicationIds]
 * @property {string} [comment]
 * @property {Email} [email]
 * @property {boolean} [emailedUser]
 * @property {number} [expiry]
 * @property {string} [localizedAction]
 * @property {string} [localizedDuration]
 * @property {string} [localizedOption]
 * @property {string} [localizedReason]
 * @property {boolean} [notifyUser]
 * @property {string} [option]
 * @property {UserActionPhase} [phase]
 * @property {string} [reason]
 * @property {string} [reasonCode]
 */


/**
 * A log for an action that was taken on a User.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionLog
 *
 * @property {UUIDString} [actioneeUserId]
 * @property {UUIDString} [actionerUserId]
 * @property {Array<UUIDString>} [applicationIds]
 * @property {string} [comment]
 * @property {boolean} [emailUserOnEnd]
 * @property {boolean} [endEventSent]
 * @property {number} [expiry]
 * @property {LogHistory} [history]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {string} [localizedName]
 * @property {string} [localizedOption]
 * @property {string} [localizedReason]
 * @property {string} [name]
 * @property {boolean} [notifyUserOnEnd]
 * @property {string} [option]
 * @property {string} [reason]
 * @property {string} [reasonCode]
 * @property {UUIDString} [userActionId]
 */


/**
 * Models content user action options.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionOption
 *
 * @property {LocalizedStrings} [localizedNames]
 * @property {string} [name]
 */


/**
 * The phases of a time-based user action.
 *
 * @author Brian Pontarelli
 *
 * @readonly
 * @enum
 */
var UserActionPhase = {
  start: "start",
  modify: "modify",
  cancel: "cancel",
  end: "end"
};

/**
 * Models action reasons.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionReason
 *
 * @property {string} [code]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {LocalizedStrings} [localizedTexts]
 * @property {string} [text]
 */


/**
 * User Action Reason API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionReasonRequest
 *
 * @property {UserActionReason} [userActionReason]
 */


/**
 * User Action Reason API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionReasonResponse
 *
 * @property {UserActionReason} [userActionReason]
 * @property {Array<UserActionReason>} [userActionReasons]
 */


/**
 * User Action API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionRequest
 *
 * @property {UserAction} [userAction]
 */


/**
 * User Action API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserActionResponse
 *
 * @property {UserAction} [userAction]
 * @property {Array<UserAction>} [userActions]
 */


/**
 * Models the User Bulk Create Event (and can be converted to JSON).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserBulkCreateEvent
 * @extends BaseEvent
 *
 * @property {Array<User>} [users]
 */


/**
 * A log for an event that happened to a User.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserComment
 *
 * @property {string} [comment]
 * @property {UUIDString} [commenterId]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {UUIDString} [userId]
 */


/**
 * @author Seth Musselman
 *
 * @typedef {Object} UserCommentRequest
 *
 * @property {UserComment} [userComment]
 */


/**
 * User Comment Response
 *
 * @author Seth Musselman
 *
 * @typedef {Object} UserCommentResponse
 *
 * @property {UserComment} [userComment]
 * @property {Array<UserComment>} [userComments]
 */


/**
 * Models a User consent.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserConsent
 *
 * @property {Consent} [consent]
 * @property {UUIDString} [consentId]
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [giverUserId]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {ConsentStatus} [status]
 * @property {UUIDString} [userId]
 * @property {Array<string>} [values]
 */


/**
 * API response for User consent.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserConsentRequest
 *
 * @property {UserConsent} [userConsent]
 */


/**
 * API response for User consent.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserConsentResponse
 *
 * @property {UserConsent} [userConsent]
 * @property {Array<UserConsent>} [userConsents]
 */


/**
 * Models the User Create Event (and can be converted to JSON).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserCreateEvent
 * @extends BaseEvent
 *
 * @property {User} [user]
 */


/**
 * Models the User Deactivate Event (and can be converted to JSON).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserDeactivateEvent
 * @extends BaseEvent
 *
 * @property {User} [user]
 */


/**
 * Models the User Event (and can be converted to JSON) that is used for all user modifications (create, update,
 * delete).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserDeleteEvent
 * @extends BaseEvent
 *
 * @property {User} [user]
 */


/**
 * User API delete request object.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserDeleteRequest
 *
 * @property {boolean} [dryRun]
 * @property {boolean} [hardDelete]
 * @property {string} [query]
 * @property {string} [queryString]
 * @property {Array<UUIDString>} [userIds]
 */


/**
 * User API bulk response object.
 *
 * @author Trevor Smith
 *
 * @typedef {Object} UserDeleteResponse
 *
 * @property {boolean} [dryRun]
 * @property {boolean} [hardDelete]
 * @property {number} [total]
 * @property {Array<string>} [userIds]
 */


/**
 * Models the User Email Verify Event (and can be converted to JSON).
 *
 * @author Trevor Smith
 *
 * @typedef {Object} UserEmailVerifiedEvent
 * @extends BaseEvent
 *
 * @property {User} [user]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserinfoResponse
 * @extends Object<string, Object>
 *
 */


/**
 * Models the User Login Failed Event.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserLoginFailedEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [authenticationType]
 * @property {string} [ipAddress]
 * @property {User} [user]
 */


/**
 * Models the User Login Success Event.
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserLoginSuccessEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [authenticationType]
 * @property {UUIDString} [connectorId]
 * @property {UUIDString} [identityProviderId]
 * @property {string} [identityProviderName]
 * @property {string} [ipAddress]
 * @property {User} [user]
 */


/**
 * @typedef {Object} UsernameModeration
 * @extends Enableable
 *
 * @property {UUIDString} [applicationId]
 */


/**
 * Models the User Password Breach Event.
 *
 * @author Matthew Altman
 *
 * @typedef {Object} UserPasswordBreachEvent
 * @extends BaseEvent
 *
 * @property {User} [user]
 */


/**
 * Models the User Reactivate Event (and can be converted to JSON).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserReactivateEvent
 * @extends BaseEvent
 *
 * @property {User} [user]
 */


/**
 * User registration information for a single application.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserRegistration
 *
 * @property {UUIDString} [applicationId]
 * @property {string} [authenticationToken]
 * @property {UUIDString} [cleanSpeakId]
 * @property {Object<string, Object>} [data]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastLoginInstant]
 * @property {number} [lastUpdateInstant]
 * @property {Array<string>} [preferredLanguages]
 * @property {Array<string>} [roles]
 * @property {string} [timezone]
 * @property {Object<string, string>} [tokens]
 * @property {string} [username]
 * @property {ContentStatus} [usernameStatus]
 * @property {boolean} [verified]
 */


/**
 * Models the User Create Registration Event (and can be converted to JSON).
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserRegistrationCreateEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {UserRegistration} [registration]
 * @property {User} [user]
 */


/**
 * Models the User Delete Registration Event (and can be converted to JSON).
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserRegistrationDeleteEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {UserRegistration} [registration]
 * @property {User} [user]
 */


/**
 * Models the User Update Registration Event (and can be converted to JSON).
 *
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserRegistrationUpdateEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {UserRegistration} [original]
 * @property {UserRegistration} [registration]
 * @property {User} [user]
 */


/**
 * Models the User Registration Verified Event (and can be converted to JSON).
 *
 * @author Trevor Smith
 *
 * @typedef {Object} UserRegistrationVerifiedEvent
 * @extends BaseEvent
 *
 * @property {UUIDString} [applicationId]
 * @property {UserRegistration} [registration]
 * @property {User} [user]
 */


/**
 * User API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserRequest
 *
 * @property {boolean} [sendSetPasswordEmail]
 * @property {boolean} [skipVerification]
 * @property {User} [user]
 */


/**
 * User API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserResponse
 *
 * @property {string} [emailVerificationId]
 * @property {Object<UUIDString, string>} [registrationVerificationIds]
 * @property {string} [token]
 * @property {User} [user]
 */


/**
 * This class is the user query. It provides a build pattern as well as public fields for use on forms and in actions.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserSearchCriteria
 * @extends BaseElasticSearchCriteria
 *
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var UserState = {
  Authenticated: "Authenticated",
  AuthenticatedNotRegistered: "AuthenticatedNotRegistered",
  AuthenticatedNotVerified: "AuthenticatedNotVerified",
  AuthenticatedRegistrationNotVerified: "AuthenticatedRegistrationNotVerified"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} UserTwoFactorConfiguration
 *
 * @property {Array<TwoFactorMethod>} [methods]
 * @property {Array<string>} [recoveryCodes]
 */


/**
 * Models the User Update Event (and can be converted to JSON).
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} UserUpdateEvent
 * @extends BaseEvent
 *
 * @property {User} [original]
 * @property {User} [user]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} ValidateResponse
 *
 * @property {JWT | object} [jwt]
 */


/**
 * @author Daniel DeGroff
 *
 * @readonly
 * @enum
 */
var VerificationStrategy = {
  ClickableLink: "ClickableLink",
  FormField: "FormField"
};

/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} VerifyEmailRequest
 *
 * @property {string} [oneTimeCode]
 * @property {string} [verificationId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} VerifyEmailResponse
 *
 * @property {string} [oneTimeCode]
 * @property {string} [verificationId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} VerifyRegistrationRequest
 *
 * @property {string} [oneTimeCode]
 * @property {string} [verificationId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} VerifyRegistrationResponse
 *
 * @property {string} [oneTimeCode]
 * @property {string} [verificationId]
 */


/**
 * @author Daniel DeGroff
 *
 * @typedef {Object} VersionResponse
 *
 * @property {string} [version]
 */


/**
 * A server where events are sent. This includes user action events and any other events sent by FusionAuth.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} Webhook
 *
 * @property {Array<UUIDString>} [applicationIds]
 * @property {number} [connectTimeout]
 * @property {Object<string, Object>} [data]
 * @property {string} [description]
 * @property {Object<EventType, boolean>} [eventsEnabled]
 * @property {boolean} [global]
 * @property {HTTPHeaders} [headers]
 * @property {string} [httpAuthenticationPassword]
 * @property {string} [httpAuthenticationUsername]
 * @property {UUIDString} [id]
 * @property {number} [insertInstant]
 * @property {number} [lastUpdateInstant]
 * @property {number} [readTimeout]
 * @property {string} [sslCertificate]
 * @property {string} [url]
 */


/**
 * Webhook API request object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} WebhookRequest
 *
 * @property {Webhook} [webhook]
 */


/**
 * Webhook API response object.
 *
 * @author Brian Pontarelli
 *
 * @typedef {Object} WebhookResponse
 *
 * @property {Webhook} [webhook]
 * @property {Array<Webhook>} [webhooks]
 */


/**
 * @author Brett Pontarelli
 *
 * @typedef {Object} XboxApplicationConfiguration
 * @extends BaseIdentityProviderApplicationConfiguration
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * Xbox gaming login provider.
 *
 * @author Brett Pontarelli
 *
 * @typedef {Object} XboxIdentityProvider
 * @extends BaseIdentityProvider<XboxApplicationConfiguration>
 *
 * @property {string} [buttonText]
 * @property {string} [client_id]
 * @property {string} [client_secret]
 * @property {string} [scope]
 */


/**
 * @readonly
 * @enum
 */
var XMLSignatureLocation = {
  Assertion: "Assertion",
  Response: "Response"
};

